DOCS = {
    "define_units":r"""Define how initial data is used to set the reference frequency, $\omega$ and energy scale, $\mu$.
    
    * energy scale: $M_{\rm P}$ in Planck units
    * frequency scale: initial Hubble rate, $H_0$ (in Planck units)
    """,

    "initial_conditions":"""Define how to create an array of initial data to solve the ODE's.

    Parameters
    ----------
    vals :  BGSystem
        contains initial data
    ntr : int
        truncation number

    Returns
    -------
    yini : NDArray
        initial data in an array
    """,

    "update_values" : """Compute static variables.

    Parameters
    ----------
    t : float
        time
    y : float
        array of data
    vals :  BGSystem
        system of data storing static variables
    atol, rtol : float
        precisions (used for heaviside functions)
    """,

    "compute_timestep" : """Compute time derivatives for dynamical variables.

    Parameters
    ----------
    t : float
        time
    y : float
        array of data
    vals :  BGSystem
        system of data storing static variables
    atol, rtol : float
        precisions (used for heaviside functions)

    Returns
    -------
    dydt : NDArray
        time derivative of y
    """,

    "condition_EndOfInflation":"""
    Determine the end of inflation using $\ddot{a} = 0$

    Parameters
    ----------
    t : float
        time
    y : float
        array of data
    vals :  BGSystem
        system of data storing static variables
    """,

    "consequence_EndOfInflation":r"""
    Define how the solver response to a (non)-occurrence of the end of inflation.

    If the event occurs, the solultion is accepted. Else, increase $t_{\rm end}$ and continue solving.

    Parameters
    ----------
    vals :  BGSystem
        solution to an ODE
    occurrence : bool
        if the event has occurred during the ODE solution
    """,

    "condition_NegativeEnergies":r"""
    Determine if either $\langle {\bf E}^2 \rangle$ or $\langle {\bf B}^2 \rangle$ is negative.

    Parameters
    ----------
    t : float
        time
    y : float
        array of data
    vals :  BGSystem
        system of data storing static variables
    """
}


DOCS = {
    "module": r"""
    This module is used to compute and analyse spectra of gauge-field modes.

    Throughout this module, any gauge-field mode functions are represented by dimensionless variables

    $$\sqrt{2k} A_\lambda(t,k), \quad \sqrt{\frac{2}{k}}\, a(t)\dot{A}_\lambda(k, t), \quad \lambda = \pm 1$$

    where $A_\lambda(t,k)$ are the mode functions of helicity $\lambda$ and momentum $k$ for a canonically quantized Abelian gauge-field $A_\mu(t, {\bf x})$ in Coulomb & Weyl gauge.
    The momentum variables $k$ are always returned in numerical units, i.e., $\bar{k} = k/\omega$.

    The class `BaseModeSolver` is designed to solve the second order mode equation

    $$\ddot{A}_\lambda(t,k) + P_\lambda(t,k)\dot{A}_\lambda(t,k) + Q_\lambda(t,k)A_\lambda(t,k) = 0 \, .$$

    The base class in particular is set to solve the mode equation of pure axion inflation,
    $$P_\lambda(t,k) = H \, \qquad Q_\lambda(t,k) = \left(\frac{k}{a}\right)^2  - 2\lambda \left(\frac{k}{a}\right) \xi H \, ,$$
    with Hubble rate $H$, scale factor $a$ and instability parameter $\xi$.

    To create a mode solver with custom values for $P(t,k)$ and $Q(t,k)$, use the class factory `ModeSolver`.

    The module also contains a class `GaugeSpec` designed for directly working on the spectrum of modes $A_\lambda(t,k)$.
    In particular, it is used to integrate the spectrum to obtain the quantities

    $$ \mathcal{F}_\mathcal{E}^{(n)}(t) = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{a^2 k^{n+3}}{2 \pi^2 k_{{\rm UV}}^{n+4}}  \sum_{\lambda}\lambda^n |\dot{A}_\lambda(t,k)|^2\,,$$
    $$ \mathcal{F}_\mathcal{G}^{(n)}(t) = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{a k^{n+4}}{2 \pi^2 k_{{\rm UV}}^{n+4}}\sum_{\lambda}\lambda^{n+1} \operatorname{Re}[\dot{A}_\lambda(t,k)A_\lambda^*(t,k)]\,,$$
    $$ \mathcal{F}_\mathcal{B}^{(n)}(t) = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{k^{n+5}}{2 \pi^{2}k_{{\rm UV}}^{n+4}} \sum_{\lambda}\lambda^n |A_\lambda(t,k)|^2\,,$$

    which may be used to estimate the error of a GEF solution.
    """,

    "GaugeSpec":r"""
    A class representing a spectrum of gauge-field modes as a function of time.

    This class inherits from `dict` and needs the following keys:  
    't', 'N', 'k', 'Ap', 'dAp', 'Am', 'dAm'

    All quantities are represented in numerical units (see `GEFF.bgtypes.BGSystem`).

    The spectrum can be evaluated at certain times $t$ or for certain momenta $k$ by using `tslice` and `kslice`
    Furthermore, the spectrum contained in the object can be integrated to compute gauge-field expectation values.
     The result can be used to estimate the error of a GEF run.

    Attributes
    ----------
    t : NDArray
        the cosmic time coordinates $t$ of the spectrum
    N : NDArray
        the $e$-folds as a function of cosmic time, $N(t)$
    k : NDarray
        the momenta $k$ at which the spectrum is evaluated
    Ap, Am : NDarray
        the mode functions, $\sqrt{2 k} A_\pm(k, t)$
    dAp, dAm : NDarray
        the mode-function derivatives, $\sqrt{2/k} \, e^{N(t)}\dot{A}_\pm(k, t)$
    """,

    "GaugeSpec.estimate_GEF_error": r"""
        Estimate the relative deviation between a GEF solution and the mode spectrum by computing

        $$\varepsilon_\mathcal{X} = \left|1 - \frac{\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm MbM}}{\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm GEF}}\right|$$

        for $\mathcal{X} = \mathcal{E},\,\mathcal{B},\,\mathcal{G}$. Here, $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm MbM}$ are the integrals computed by `integrate`, 
        $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm GEF}$ refer to the same quantity in `BG`.
        If the time coordinate of `BG` does not align with the spectrum, its values are interpolated.

        Because $k_{\rm UV}(t)$ increases monotonically, the spectrum contains only few relevant modes $k < k_{\rm UV}(t)$ at early times.
        This poses a problem for the numerical integration of $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm MbM}$.
        To avoid claiming a disagreement between $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm MbM}$ and $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm GEF}$ due to this effect,
        errors with $\varepsilon_\mathcal{X} > \varepsilon_{\rm thr}$ are discarded until the first time when $\varepsilon_\mathcal{X} < \varepsilon_{\rm thr}$.

        As the integration result fluctuates significantly for few momenta $k < k_{\rm UV}(t)$ when using `simpson`,
        the errors can be binned by setting `binning`. The reported error is the average over a bin of width $(t_{i}, t_{i+\Delta})$ with $\Delta$ set by `binning`.
        This binned error is then associated to the time $(t_{i} + t_{i+\Delta})/2$. For `quad`, `binning` can also be set to `None`.
        For details on the integration methods `simpson` and `quad`, see `SpecSlice.integrate_slice`.

        Parameters
        ----------
        BG : BGSystem
            the system where $\big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm GEF}$ are stored
        references : list of str
            the names where $(k_{\rm UV}/a)^4 \big(\mathcal{F}_\mathcal{X}^{(0)}\big)_{\rm GEF}$ are stored in `BG`
        cutoff : str
            the name where the UV-cutoff, $k_{\rm UV}$, is stored in `BG`
        err_thr : float
            the error threshold $\varepsilon_{\rm thr}$
        binning : int or None
            the bin size $\Delta$ (no binning if `None`)
        verbose : bool
            if `True`, print a summary of the errors
        **IntegratorKwargs :  kwargs
            passed to `SpecSlice.integrate_slice`

        Returns
        -------
        errs : list of NDArray
            a list of the binned errors with entries $[\varepsilon_\mathcal{E},\varepsilon_\mathcal{B}, \varepsilon_\mathcal{G}]$
        terr : NDArray
            the time coordinates corresponding to `errs`
        og_errs : list of NDArray
            the same as `errs` but without binning
        """,

    "SpecSlice": r"""
    A class representing a spectrum of gauge-field modes at a time $t$.

    Instances of this class are created by `GaugeSpec.tslice`. The main purpose of this class is to integrate the spectrum at time $t$ using `integrate_slice`.

    Attributes
    ----------
    t : NDArray
        the cosmic time coordinates $t$ of the spectrum
    N : NDArray
        the $e$-folds as a function of cosmic time, $N(t)$
    k : NDarray
        the momenta $k$ at which the spectrum is evaluated
    Ap, Am : NDarray
        the mode functions, $\sqrt{2 k} A_\pm(k, t)$
    dAp, dAm : NDarray
        the mode-function derivatives, $\sqrt{2/k} \, e^{N(t)}\dot{A}_\pm(k, t)$
    """,

    "SpecSlice.integrate_slice":r"""
        Compute the three integrals $\mathcal{F}_\mathcal{X}^{(n)}(t)$ for $\mathcal{X} = \mathcal{E}, \mathcal{B},\mathcal{G}$ for a fixed time $t$ and index $n$.

        The integrals can either be computed directly using `simpson` or `quad` from `scipy.interpolate`. When using `quad` the data for $\sqrt{2 k} A_\pm(k, t)$, $\sqrt{2/k} \, e^{N(t)}\dot{A}_\pm(k, t)$
          are interpolated to obtain smooth functions. To avoid this, it is recommended to use `simpson`.

        When using `simpson`, the integral is only computed if $m > m_{\rm thr}$ momenta $k_i$ satisfy $k < k_{\rm UV}$. Otherwise, the integral is set to zero.

        When using `quad`, the absolute and relative tolerances of the integrator are set by `epsabs` and `epsrel`. The interpolation method is defined by `interpolator`.
        Currently, only `CubicSpline` and `PchipInterpolator` from `scipy.interpolate` are supported. The later is preferred as interpolating the oscillatory mode functions can be subject to "overshooting".
        See [scipy's tutorial for 1-D interpolation](https://docs.scipy.org/doc/scipy/tutorial/interpolate/1D.html#tutorial-interpolate-1dsection) for more details.

        Parameters
        ----------
        n : int
            the integer $n$ in $\mathcal{F}_\mathcal{X}^{(n)}(t)$ for $\mathcal{X} = \mathcal{E}, \mathcal{B},\mathcal{G}$
        integrator : str
            set the integration method to `simpson` or `quad`
        modethr : int
            set $m_{\rm thr}$ when using `simpson`
        epsabs : float
            the absolute tolerance of `quad`
        epsrel : float
            the relative tolerance of `quad`  
        interpolator
            the interpolator used to get smooth functions for `quad`
        

        Returns
        -------
        FMbM : NDArray
            contains [$\mathcal{F}_\mathcal{E}^{(n)}(t)$, $\mathcal{F}_\mathcal{B}^{(n)}(t)$, $\mathcal{F}_\mathcal{B}^{(n)}(t)$] and the error estimated by `quad`.
             When using `simpson` the error is a dummy output. The shape of the result is (3, 2) with the second index indicating the integral (at 0), or the error (at 1).
        """,

    "BaseModeSolver":r"""
    A class used to compute gauge-field modes evolving on a time-dependent background.

    This class is used to evolve the gauge-field modes $A_\pm(t,k)$ as defined by `mode_equation`.
     To do so, the evolution of the time-dependent background is obtained from a GEF solution.

    The modes are initialized deep inside the Bunch&ndash;Davies vacuum as given by `initialise_in_bd`. 
    Numerically, the initialization time is implicitly defined by the condition $k = 10^{5/2} k_{\rm UV}(t_{\rm ini})$, with $k_{\rm UV}(t)$ obtained from the GEF solution.
    At times $t < t_{\rm ini}$ the mode is assumed to be in Bunch&ndash;Davies.

    The mode equations are solved with an explicit Runge&ndash;Kutta of order 5(4), which is implemented in `scipy.integrate.solve_ivp`.

    For creating a custom subclass of `BaseModeSolver` with user-specified  mode equation and initial conditions, you can use the class factory `ModeSolver`.
    """,

    "ModeSolver":r"""
    Create a subclass of `BaseModeSolver` with custom mode equation and initial conditions.

    In case your GEF model does not follow the pre-defined gauge-field mode equation `BaseModeSolver.mode_equation`,
    or initial conditions, `BaseModeSolver.initialise_in_bd` this method  defines a new subclass with
     these methods replaced by `new_mode_eq` and `new_bd_init`.
    
    The method `new_mode_eq` needs to obey the following rules:
    1. The call signature is `f(t,y,k,**kwargs)`
    2. The arguments `t` / `k` expect floats representing time / momentum
    3. The argument `y` expects a `numpy.ndarrray` of shape (8,) with indices
        -  0 & 2 / 4 & 6: real & imaginary part of $\sqrt{2k} A_\lambda(t,k)$ for $\lambda = 1 \, / -1$
        -  1 & 3 / 5 & 7: real & imaginary part of $\sqrt{2/k}\, a \dot{A}_\lambda(t,k)$ for $\lambda = 1 \, / -1$
    4. The kwargs are functions of the argument `t`.
    5. The return is the time derivative of `y`

    The method `new_bd_init` needs to obey the following rules:
    1. The call signature is `f(t,k,**kwargs)`
    2. The arguments `t` / `k` expect floats representing time / momentum
    3. The kwargs are functions of the argument `t`.
    4. The return is a `numpy.ndarrray` of shape (8,)  with indices
        -  0 & 2 / 4 & 6: real & imaginary part of $\sqrt{2k} A_\lambda(t_{\rm init},k)$ for $\lambda = 1 \, / -1$
        -  1 & 3 / 5 & 7: real & imaginary part of $\sqrt{2/k}\, a \dot{A}_\lambda(t_{\rm init},k)$ for $\lambda = 1 \, / -1$
    
    The lists `ode_keys` and `init_keys` are handled as follows:
    - `ode_keys` and `init_keys` need to contain the keys associated to the respective kwargs of `new_mode_eq` and `new_bd_init`.
    - These keys correspond to names of `GEFF.bgtypes.Val` objects belonging to a `GEFF.bgtypes.BGSystem` passed to the class upon initialisation.
        The respective `Val` objects are interpolated to obtain functions of time. 
        These functions are then passed to to the corresponding keyword arguments of `new_mode_eq`  and `new_bd_init`.
    - `ode_keys` and `init_keys` are added to `BaseModeSolver.necessary_keys` of the new subclass.

    The `BaseModeSolver.cutoff` and `BaseModeSolver.atol` attributes can also be adjusted.

    Parameters
    ----------
    new_mode_eq : Callable
        a new mode equation
    ode_keys : list of str
        the non-standard keywords of `new_mode_eq`
    new_bd_init : Callable
        a new mode bd initial condition
    init_keys : list of str
        the non-standard keywords of `new_bd_init`
    new_cutoff : str
        the new `cutoff` attribute of the subclass
    default_atol : float
        the default absolute tolerance used by the subclass

    Returns
    -------
    NewModeSolver
        the newly defined subclass of `BaseModeSolver`

    Example
    -------
    ```python
        import numpy as np
        from GEFF.bgtypes import BGSystem, define_var, t, N, kh

        # Define a new mode equation:
        def custom_mode_eq(t, y, k, a, X, Y):
            #create a return array of the right shape
            dydt = np.ones_like(y)

            #compute real-part time derivatives for positive modes
            dydt[0] = k / a(t) * y[1] # a is a function of t.
            dydt[1] = X(t)/Y(t)*y[0] # X and Y are functions of t.

            #compute imag-part time derivatives for positive modes
            ...

            #compute real-part time derivatives for negative modes
            ...
            ...

            return dydt

        # Define a new initial condition for the modes:
        def custom_bd_init(t, k, alpha):
            y = alpha(t)*np.array([...]) # alpha is a function of t.
            return y

        # the kwargs of custom_mode_eq are 'a', 'X' and 'Y':
        custom_ode_keys = ['a', 'X', 'Y']

        # the kwarg of custom_bd_init is 'alpha':
        custom_init_keys = ['alpha']

        # Define the custom mode solver using the class factory:
        CustomModeSolver = ModeSolver(custom_mode_eq, custom_ode_keys,
                                         custom_bd_init, custom_init_keys)

        # To initialise CustomModeSolver we need a BGSystem. 
        # Its Variables need to have the right names however:
        # The default: 't', 'N', 'kh' were loaded from GEFF.bgtypes

        # Because of custom_mode_eq we also need 'a', 'X', 'Y'
        a = define_var("a", 0, 0)
        X = define_var("X", 2, 0)
        Y = define_var("Y", 2, 0)

        # For custom_bd_init we need 'alpha'
        alpha = define_var("alpha", 0, 0)

        # When in doubt, consult necessary_keys:
        print(CustomModeSolver.necessary_keys)

        # We create the BGSystem and initialise all its values:
        sys = BGSystem({t, N, kh, a, X, Y, alpha}, 1e-6, 1)
        sys.initialise("t")(...)
        ...

        # The values in sys can now be used to initialise CustomModeSolver
        MbM = CustomModeSolver(sys)

        # Let's compute a spectrum using the new setup:
        spec = MbM.compute_spectrum(100)
        ...
        ```
    """
}


from ._docs import generate_docs, docs_bgtypes
import numpy as np
from typing import Callable, ClassVar
import pandas as pd
from copy import deepcopy

class BGSystem:
    def __init__(self, quantity_set : set, omega : float, mu : float):
        """
        Create a new BGSystem in physical units.

        Parameters
        ----------
        quantity_set : set of Quantity
            used to define `quantities`
        omega : float
            the characteristic frequency scale
        mu : float
            the characteristic energy scale
        """
        self.quantities : dict = {q.name:q for q in quantity_set}
        """A dictionary of all `Quantity` objects for this BGSystem"""
        self._omega : float = omega
        self._mu : float = mu
        self._units=True

    @property
    def omega(self) -> float:
        """A frequency scale (typically the Hubble rate at some reference time)"""
        return self._omega
    
    @property 
    def mu(self) -> float:
        """An energy scale (typically the Planck mass)"""
        return self._mu

    @property
    def units(self) -> bool:
        """Indicates the current units of the BGSystem. `True`:physical units, `False`:numerical units"""
        return self._units
    
    @units.setter
    def units(self, newunits:bool):
        """Change the units of the BGSystem and its `Quantity` instances."""
        for var in vars(self):
            obj = getattr(self, var)
            if isinstance(obj, Quantity):
                obj.units = newunits
        self._units=bool(newunits)
        return

    
    @classmethod
    def from_system(cls, sys : 'BGSystem', copy : bool=False) -> 'BGSystem':
        """Initialize a new `BGSystem` from an existing instance.

        The new instance is created with the same quantities, reference frequency and amplitude as the original.
        If specified, the `Quantity` instances are also copied to the new BGSystem

        Parameters
        ----------
        sys : BGSystem
            the original instance used as a template
        copy : Boolean
            `True` if `Quantity` instances are also copied
        Returns
        -------
        newinstance : BGSystem
            the new instance
        """

        newinstance = cls(sys.quantity_set(), sys.omega, sys.mu)

        if copy:
            #store units of original sys
            units = sys.units
            #match units of new system
            sys.units = True

            #Copy values and functions
            values = sys.variable_list()
            funcs = sys.function_list()
            consts = sys.constant_list()

            for const in consts:
                obj = const
                newinstance.initialise(obj.name)(obj.value)

            for value in values:
                obj = value
                newinstance.initialise(obj.name)(obj.value)

            for func in funcs:
                obj = func
                newinstance.initialise(obj.name)(obj.basefunc)
            
            #restore old units
            sys.units = units
        
        return newinstance
    
    def quantity_set(self) -> set[object]:
        """
        Get a set of all `Quantity` objects attributed to this BGSystem.

        Returns
        -------
        set : set
            a set of objects.
        """

        return set(self.quantities.values())
    
    def quantity_names(self) -> list[str]:
        """
        Get a list of names for all `Quantity` objects attributed to this BGSystem.

        Returns
        -------
        names : list of str
            the list of names.
        """

        return list(self.quantities.keys())
    

    def initialise(self, quantity : str) -> Callable:
        """
        Instantiate a `Quantity` object from `quantities`.

        The method creates a function `init` which can be called by
        an arithmetic type / `Callable` to instantiate a `Val` / `Func`.
        Calling `init` adds and instance of the `Quantity` as a new attribute to the BGSystem,
        with the attribute name corresponding to the object's `name` attribute.   

        Parameters
        ----------
        quantity : str
            the name of the object which is to be instantiated.

        Returns
        -------
        init : Callable
            a function used to initialize the `Quantity` object
        """

        def init(obj : np.ndarray | Callable):
            """
            Initialize a `Quantity` object with an arithmetic type / Callable.

            This adds an instance of the `Quantity` as a new attribute to the `BGSystem`, with the attribute name
            corresponding to the `Quantity` object's `name` attribute. 

            Parameters
            ----------
            obj : NDArray or Callable
                the NDArray / Callable with which the `Quantity` is to be instantiated.
            """

            q = self.quantities[quantity]
            setattr( self, quantity, q(obj, self) )
            return
        
        return init
    
    def variable_list(self) -> list['Variable']:
        """
        Get a list of all `Variable` instances attributed to this BGSystem.

        Returns
        -------
        vals : list of Variable
            the list of `Variable` instances.
        """

        vals = []
        for var in vars(self):
            obj = getattr(self, var)
            if isinstance(obj, Variable):
                vals.append(obj)    
        return vals
    
    def variable_names(self) -> list[str]:
        """
        Get a list of names for all `Variable` instances attributed to this BGSystem.

        Returns
        -------
        names : list of str
            the list of names.
        """

        names = []
        for val in self.variable_list():
            names.append(val.name)
        return names
    
    def constant_list(self) -> list['Constant']:
        """
        Get a list of all `Constant` instances attributed to this BGSystem.

        Returns
        -------
        vals : list of Val
            the list of `Constant` instances.
        """

        vals = []
        for var in vars(self):
            obj = getattr(self, var)
            if isinstance(obj, Constant):
                vals.append(obj)    
        return vals
    
    def constant_names(self) -> list[str]:
        """
        Get a list of names for all `Constant` instances attributed to this BGSystem.

        Returns
        -------
        names : list of str
            the list of names.
        """

        names = []
        for val in self.constant_list():
            names.append(val.name)
        return names

    def function_list(self) -> list['Func']:
        """
        Get a list of all `Func` instances attributed to this BGSystem.

        Returns
        -------
        funcs : list of Func
            the list of `Func` instances.
        """

        funcs = []
        for var in vars(self):
            obj = getattr(self, var)
            if isinstance(obj, Func):
                    funcs.append(obj)      
        return funcs
    
    def function_names(self) -> list[str]:
        """
        Get a list of names for all `Func` instances attributed to this BGSystem.

        Returns
        -------
        names : list of str
            the list of names.
        """

        names = []
        for val in self.function_list():
            names.append(val.name)
        return names
    
    def remove(self, name : str):
        """
        Remove a `Quantity` object and its instance from the BGSystem.

        Parameters
        ----------
        name : str
            the name of the object
        """

        delattr(self, name)
        self.quantities.pop(name)
        return
    
    def add_variable(self, name : str, qu_omega : int, qu_mu : int):
        """
        Define a new `Variable` object and add it to `quantities`.

        Parameters
        ----------
        name : str
            the name of the new object.
        qu_omega : int
            the 'u_omega' parameter of the new object.
        qu_mu : int
            the 'u_mu' parameter of the new object.
        """

        self.quantities[name] = define_var(name, qu_omega, qu_mu)
        return
    
    def add_constant(self, name : str, qu_omega : int, qu_mu : int):
        """
        Define a new `Constant` object and add it to `quantities`.

        Parameters
        ----------
        name : str
            the name of the new object.
        qu_omega : int
            the 'u_omega' parameter of the new object.
        qu_mu : int
            the 'u_mu' parameter of the new object.
        """

        self.quantities[name] = define_const(name, qu_omega, qu_mu)
        return
    
    def add_function(self, name : str, args : list['Val'], qu_omega : int, qu_mu : int):
        """
        Define a new `Func` object and add it to `quantities`.

        Parameters
        ----------
        name : str
            the name of the new object.
        args : list of BGVal
            the 'args' parameter of the new object.
        qu_omega : int
            the 'u_omega' parameter of the new object.
        qu_mu : int
            the 'u_mu' parameter of the new object.
        """

        self.quantities[name] = define_func(name, args, qu_omega, qu_mu)
        return
    
    def save_variables(self, path : str):
        """
        Save the data in the current GEF instance in an output file.

        Note, data is always stored in numerical units.
        The save will not store constants or functions, only variables.

        Parameters
        ----------
        path : str
            Path to store data in.

        Raises
        ------
        ValueError
            if the GEF object has no data to store.
        """

        storeables = self.variable_names()    
        #Check that all dynamic and derived quantities are initialised in this GEF instance
        if len(storeables) == 0:
            raise ValueError("No data to store.")
        
        #Create a dictionary used to initialise the pandas DataFrame
        dic = {}

        #remember the original units of the GEF
        og_units=self.units

        #Data is always stored unitless
        self.units = False

        for key in storeables:
            dic[key] = getattr(self, key).value
        
        #Create pandas data frame and store the dictionary under the user-specified path
        output_df = pd.DataFrame(dic)  
        output_df.to_csv(path)

        #after storing data, restore original units
        self.units = og_units
        return
    
    def load_variables(self, path : str):
        """
        Load data and store its results in the BGSystem.

        Note, data is always loaded assuming numerical units.
        Data is only loaded for variables, not for functions or constants.

        Parameters
        ----------
        path : None or str
            Path to load data from

        Raises
        ------
        FileNotFoundError
            if no file is found at `path`.
        AttributeError
            if the file contains a column labeled by a key which does not match any BGSystem variable.
        """
        #Check if file exists
        try:
            #Load from file
            input_df = pd.read_table(path, sep=",")
        except FileNotFoundError:
            raise FileNotFoundError(f"No file found under '{path}'")
        
        #Dictionary for easy access using keys

        data = dict(zip(input_df.columns[1:],input_df.values[:,1:].T))

        #Before starting to load, check that the file is compatible with the GEF setup.
        names = self.quantity_names()
        for key in data.keys():
            if key not in names:
                raise AttributeError(f"The data table you tried to load contains an unknown quantity: '{key}'")
        
        #Store current units to switch back to later
        og_units=self.units

        #GEF data is always stored untiless, thus it is assumed to be untiless when loaded.
        self.units = False
        #Load data into background-value attributes
        for key, values in data.items():
            if key in self.variable_names():
                getattr(self, key).value = values
            else: 
                self.initialise(key)(values)

        self.units = og_units
        return

    
class Quantity:
    name : ClassVar[str]= ""
    """The objects name"""
    description : ClassVar[str]= ""
    """A brief description of the object"""
    u_omega : ClassVar[int] = 0
    """Indicates how the object scales with frequency."""
    u_mu : ClassVar[int] = 0
    """Indicates how the object scales with energy."""

    def __init__(self, sys : BGSystem):
        """
        Create a new Quantity as part of a BGSystem

        The units of the new object matches those of the BGSystem

        Parameters
        ----------
        sys : BGSystem
            the BGSystem to which the object belongs
        """

        self._units = sys.units
        self._conversion = (sys.omega**self.u_omega*sys.mu**self.u_mu)

    def __repr__(self):
        r"""A string representing the class, giving its name and scaling with frequency ($\omega$) and energy ($\mu$)."""
        return f"{self.name}({self.u_omega},{self.u_mu})"

    def __str__(self) -> str:
        """The class instance as a string including its name and current units."""

        if not(self._units):
            return f"{self.name} (numerical)"
        elif self._units:
            return f"{self.name} (physical)"
    
    
    @property
    def units(self) -> bool:
        """Indicates the current units of the Quantity. `True`:physical units, `False`:numerical units"""
        return self._units
    
    @units.setter
    def units(self, units : bool):
        """Convert the object between numerical and physical units."""
        self._units = bool(units)
        return
    
    @property
    def conversion(self) -> float:
        """A conversion factor between numerical and physical units."""
        return self._conversion
    
    @classmethod
    def get_description(cls) -> str:
        """Return a string describing the object."""
        if cls.description=="":
            return f"{cls.name}"
        else:
            return f"{cls.name} - {cls.description}"
        

    
class Val(Quantity):
    def __init__(self, value : np.ndarray|float, sys : BGSystem):
        """
        Create a new instance using a BGSystem.

        Parameters
        ----------
        value : NDArray
            the underlying array of the instance
        sys : BGSystem
            the BGSystem to which the instance belongs
        """
        super().__init__(sys)
        self._value =  value*self._conversion**(-sys.units)

    def __str__(self) -> str:
        """
        The class instance as a string including its name, current units, and its value.

        Returns
        -------
        str
            the string representation.
        """

        if not(self._units):
            return f"{self.name} (numerical): {self.value}"
        elif self._units:
            return f"{self.name} (physical): {self.value}"
    
    @property
    def value(self) -> float:
        """The objects value in its respective units."""
        return self._value*self._conversion**self._units
    
    @value.setter
    def value(self, newval):
        """Overwrite the `value` attribute (assuming its current units)."""
        self._value = newval*self._conversion**(-self._units)
        return
    
    #Define all mathematical operations as operations acting on self.value
    def __abs__(self):
        return abs(self.value)
    
    def __neg__(self):
        return -self.value
    
    def __pos__(self):
        return +self.value
    
    def __add__(self, other):
        #self.__Compatible(self, other, "+")
        return self.value + other
        
    __radd__ = __add__
    
    def __sub__(self, other):
        #self.__Compatible(self, other, "-")
        return self.value - other
        
    def __rsub__(self, other):
        return other - self.value

    def __mul__(self, other):
        return self.value * other
        
    __rmul__ = __mul__
    
    def __floordiv__(self, other):
        return self.value // other
        
    def __rfloordiv__(self, other):
        return other // self.value
    
    def __truediv__(self, other):
        return self.value / other
        
    def __rtruediv__(self, other):
        return other / self.value
    
    def __mod__(self, other):
        return self.value % other
    
    def __pow__(self, other):
        #BGVal should never be exponentiated by another BGVal
        return self.value ** other
    
    def __eq__(self, other):
        #self.__Compatible(self, other, "==")
        return self.value == other
            
    def __ne__(self, other):
        #self.__Compatible(self, other, "!=")
        return self.value != other
    
    def __lt__(self, other):
        #self.__Compatible(self, other, "<")
        return self.value < other

    def __gt__(self, other):
        #self.__Compatible(self, other, ">")
        return self.value > other

    def __le__(self, other):
        #self.__Compatible(self, other, "<=")
        return  self.value <= other

    def __ge__(self, other):
        #self.__Compatible(self, other, ">=")
        return  self.value >= other
    

class Func(Quantity):
    args : ClassVar[list[Val]] = []
    """Indicates the argument signature for the class."""
    dtype : ClassVar[np.floating] = np.float64
    """The data type returned by `__call__`."""

    def __init__(self, func : Callable, sys : BGSystem):
        """
        Create a new instance using a `Callable` and a BGSystem

        Parameters
        ----------
        func : NDArray
            the underlying function `f(*args)` of the instance
        sys : BGSystem
            the BGSystem to which the instance belongs

        Raises
        ------
        TypeError
            if the number of arguments of `func` do not match `args`
        ValueError
            if the return of`func` can't be converted to `dtype`
        """
        super().__init__(sys)
        func = np.vectorize(func, otypes=[self.dtype])
        
        try:
            testargs = [1.0 for arg in self.args]
            assert func(*testargs).dtype
        except TypeError:
            raise TypeError("The number of non-default arguments of 'func' needs to match 'len(self.args)'.")
        except ValueError:
            raise ValueError(f"'func' must return a single value which can be converted to '{self.dtype}'")
        
        self._basefunc = func

        self._arg_conversions = [(sys.omega**arg.u_omega*sys.mu**arg.u_mu)
                                 for arg in self.args]
        
    @property
    def basefunc(self) -> Callable:
        """The underlying function which defines the `__call__` method."""
        return self._basefunc
    
    def get_arg_conversions(self) -> list[float]:
        """
        Get a list of conversion factors for each argument of `basefunc`.

        Returns
        -------
        arg_conversions : list of float
            a list of conversion factors
        """

        return self._arg_conversions
    
    def __call__(self, *args):
        """
        Define the call method of the class as outlined in its documentation.
        """
        def float_handler(x, i):
            return x*self._arg_conversions[i]**(1-self._units)
        
        def val_handler(x, i):
            conv = x.conversion
            assert self._arg_conversions[i] == conv
            return x*conv**(1-x.units)

        typedic = {Variable : val_handler, Val : val_handler, Constant: val_handler}

        args = [typedic.get(arg.__class__.__bases__[0], float_handler)(arg, i) for i, arg in enumerate(args)]

        return self._basefunc(*args)/self._conversion**(1-self._units)

    
class Variable(Val):
    dtype : ClassVar[np.floating] = np.float64
    """The data type of `value`."""
    
    def __init__(self, value : np.ndarray, sys : BGSystem):
        """
        Create a new instance using a numpy array and a BGSystem

        Parameters
        ----------
        value : NDArray
            the underlying array of the instance
        sys : BGSystem
            the BGSystem to which the instance belongs
        """
        super().__init__( np.asarray(value, dtype=self.dtype), sys)

    #The following methods ensure that a `Val` instance can be used as an array concerning mathematical operations and indexing.
    
    def __getitem__(self, idx):
        return self.value[idx]
    
    def __len__(self):
        return len(self.value)
    
    @property
    def value(self) -> np.ndarray:
        """The objects value in its respective units."""
        return self._value*self._conversion**(self._units)

    @value.setter
    def value(self, newval):
        """Overwrite the `value` attribute ensuring it is an array"""
        self._value = np.asarray(newval*self._conversion**(-self._units), dtype=self.dtype)
        return
     
class Constant(Val):
    def __init__(self, value, sys : BGSystem):
        """
        Create a new instance using a float and a BGSystem

        Parameters
        ----------
        value : NDArray
            the underlying array of the instance
        sys : BGSystem
            the BGSystem to which the instance belongs

        Raises
        ------
        TypeError 
            if value cannot be converted to a float
        """
        super().__init__( float(value), sys)

class GaugeField:
    name : ClassVar[str] = ""
    """The name of the class."""
    zeros : ClassVar[list[Variable]]= []
    r"""A list of the associated 0$^{\rm th}$ order quantities."""
    cutoff : Variable = None
    """The associated UV cutoff scale."""
        
    @classmethod
    def get_description(cls) -> str:
        """Return a string describing the object."""
        return f"{cls.name} - associated with: {[a.name for a in cls.zeros]}, UV cutoff: {cls.cutoff.name}"
    
def define_var(qname : str, qu_omega : int, qu_mu : int, qdescription:str="", qdtype : np.dtype=np.float64):
    """
    Creates a subclass of `Variable` with custom name, and scaling.

    Parameters
    ----------
    qname : str
        the `name` attribute of the subclass
    qu_omega : int
        the `u_omega` attribute of the subclass
    qu_mu : int
        the `u_mu` attribute of the subclass
    qdescription : str
        a brief description of the subclass
    qdtype : Numpy Data Type
        the `dtype` attribute of the subclass
    
        
    Returns
    -------
    CustomVar : class
        the custom subclass

    Raises
    ------
    TypeError
        if the data type is not a subtype of `numpy.floating`
    """

    if not( np.issubdtype(qdtype, np.floating) ):
        raise TypeError("BGVal's data-type must be a subtype of 'numpy.floating'.")

    class CustomVar(Variable):
        __doc__ = docs_bgtypes.DOCS["define_var.CustomVar"]
        name=qname
        u_omega = qu_omega
        u_mu = qu_mu
        dtype = qdtype
        description = qdescription
        def __init__(self, value, sys):
            super().__init__(value, sys)
    CustomVar.__qualname__ = f"Val_{qname}"
    CustomVar.__module__ = __name__

    return CustomVar

def define_const(qname : str, qu_omega : int, qu_mu : int, qdescription:str=""):
    """
    Creates a subclass of `Constant` with custom name, and scaling.

    Parameters
    ----------
    qname : str
        the `name` attribute of the subclass
    qu_omega : int
        the `u_omega` attribute of the subclass
    qu_mu : int
        the `u_mu` attribute of the subclass
    qdescription : str
        a brief description of the subclass
        
    Returns
    -------
    CustomConst : class
        the custom subclass
    """

    class CustomConst(Constant):
        __doc__ = docs_bgtypes.DOCS["define_const.CustomConst"]
        name=qname
        u_omega = qu_omega
        u_mu = qu_mu
        description = qdescription
        def __init__(self, value, sys):
            super().__init__(value, sys)
    CustomConst.__qualname__ = f"Const_{qname}"
    CustomConst.__module__ = __name__

    return CustomConst



def define_func(qname : str, func_args : list[Val], qu_omega : int, qu_mu : int, qdescription:str="", qdtype : np.dtype=np.float64):
    """
    Creates a subclass of `Func` with custom name, scaling, and argument signature.

    Parameters
    ----------
    qname : str
        the `name` attribute of the subclass
    qu_omega : int
        the `u_omega` attribute of the subclass
    qu_mu : int
        the `u_mu` attribute of the subclass
    qdtype : Numpy Data Type
        the `dtype` attribute of the subclass
    qdescription : str
        a brief description of the subclass

    Returns
    -------
    CustomFunc : class
        the custom subclass

    Raises
    ------
    TypeError
        if the data type is not a subtype of `numpy.floating`
    """

    if not( np.issubdtype(qdtype, np.floating) ):
        raise TypeError("define_func's data-type must be a subtype of 'np.floating'.")
    
    class CustomFunc(Func):
        __doc__ = docs_bgtypes.DOCS["define_func.CustomFunc"]
        name=qname
        u_omega = qu_omega
        u_mu = qu_mu
        args = func_args
        dtype = qdtype
        description = qdescription
        def __init__(self, func, sys):
            super().__init__(func, sys)

    CustomFunc.__qualname__ = f"Func_{qname}"
    CustomFunc.__module__ = __name__

    return CustomFunc

def define_gauge(qname : str, qzeros : list[Variable], qcutoff : Variable):
    """
    A class factory creating custom  `GaugeField` classes with new name, zero variables, and cutoff scale.

    Parameters
    ----------
    qname : str
        the `name` attribute of the subclass
    qzeros : list of Variable
        the `zeros` attribute of the new subclass 
    qcutoff : Variable
        the `cutoff` attribute of the new subclass

    Returns
    -------
    CustomGaugeField : class
        the custom subclass
    """
    class CustomGaugeField(GaugeField):
        name = qname
        zeros = qzeros
        cutoff = qcutoff
        
    CustomGaugeField.__qualname__ = f"GaugeField_{qname}"
    CustomGaugeField.__module__ = __name__
    return CustomGaugeField

#Add docstrings
generate_docs(docs_bgtypes.DOCS)
    
#Some usful pre-defined quantities
#Space--time variables:
t=define_var("t", -1, 0, "cosmic time")
N=define_var("N", 0, 0, "e-folds")
a=define_var("a", 0, 0, "scale factor")
H=define_var("H", 1, 0, "Hubble rate")

#Inflaton  variables:
phi=define_var("phi", 0, 1, "inflaton expectation value")
dphi=define_var("dphi", 1, 1, "inflaton velocity")
ddphi=define_var("ddphi", 2, 1, "inflaton acceleration")

#Inflaton potential
V=define_func("V", [phi], 2, 2, "scalar potential")
dV=define_func("dV", [phi], 2, 2, "scalar-potential derivative")

#Gauge-field variables:
E=define_var("E", 4, 0, "electric-field expectation value, E^2")
B=define_var("B", 4, 0, "magnetic-field expectation value, B^2")
G=define_var("G", 4, 0, "Chern-Pontryagin expectation value, -E.B")

#Auxiliary quantities:
xi=define_var("xi", 0, 0, "instability parameter")
kh=define_var("kh", 1, 0, "instability scale")

#constants
beta=define_const("beta", 0, -1, "inflaton--gauge-field coupling beta/Mp")

#basic gauge-field
GF = define_gauge("GF", [E, B, G], kh)


import numpy as np
import inspect
from ._docs import generate_docs, docs_gef
from .bgtypes import BGSystem, Val, Func
from .models import load_model, pai

from numbers import Number
from typing import Callable

class BaseGEF:
    
    GEFSolver = pai.solver
    """The solver used to solve the GEF equations in `run`."""
    ModeSolver = pai.MbM
    """The mode solver used for mode-by-mode cross checks."""
    _input_signature = pai.model_input
    define_units = staticmethod(pai.define_units)

    def __init__(self, **kwargs):
        """
        Initialize the GEF from user-specified initial data, constants and functions.

        Parameters
        ----------
        kwargs 
            To access a list of kwargs for this model use `print_input`.

        Raises
        ------
        KeyError
            if a necessary input is missing.
        TypeError
            if the input is of the wrong type.
        """

        for val in self._input_signature:
            if val.name not in kwargs.keys():
                raise KeyError(f"Missing input '{val.name}'")
            else:
                key = val.name
                item = kwargs[key]
            if issubclass(val, Val) and not(isinstance(item, Number)):
                raise TypeError(f"Input for '{key}' should be 'Number' type.")
            elif issubclass(val, Func) and not(callable(item)):
                raise TypeError(f"Input for '{key}' must be callable.")

        #check which keys are needed for define_units
        expected_keys = inspect.getfullargspec(self.define_units).args
        omega, mu = self.define_units(**{key:kwargs[key] for key in expected_keys})

        known_objects = set().union(*[item for key, item in self.GEFSolver.known_variables.items() if key!="gauge"] )

        initial_data = BGSystem(known_objects, omega, mu)

        for key, item in kwargs.items():
            initial_data.initialise(key)(item)

        #initialise the other values with dummy variables.
        for obj in initial_data.quantity_set():
            if not(hasattr(initial_data, obj.name)):
                if issubclass(obj, Val):
                    initial_data.initialise(obj.name)(0)
                elif issubclass(obj, Func):
                    initial_data.initialise(obj.name)(lambda *x: 0)    

        self.initial_data = initial_data

        # test if ODE solver works
        try:
            self.GEFSolver(initial_data)
        except Exception:
            raise ModelError("Initializing the GEFSolver returned an error.")


    @classmethod
    def print_input(cls):
        """
        Print the input required to initialize the class.
        """
        print("This GEF model expects the following input:\n")
        for i in cls._input_signature:
            print(f" * {i.get_description()}")
        print()
        return
    
    @classmethod
    def pring_ingredients(cls):
        """
        Print a list of known variables, functions and constants for this model
        """
        print("This GEF model knows the following variables:\n")
        for key, item in cls.GEFSolver.known_variables.items():
            if len(item) > 0:
                print(f"{key}:")
                for i in item:
                    print(f" * {i.get_description()}")
        print()
        return
    
    def run(self, ntr:int=150, tend:float=120, nmodes:int=500, mbm_attempts:int=5,  resume_mbm:bool=True,
              err_tol:float=0.1, err_thr:float=0.025, binning:int=5, integrator:str="simpson", print_stats:bool=True, **solver_kwargs):
        """
        Solve the ODE's of the GEF using `GEFSolver`. Cross check the solution using `ModeSolver`.

        The `GEFSolver` is initialized using the initial conditions defined by the class.
        If the solver returns a successful solution, `ModeSolver.compute_spectrum` computes a gauge field spectrum
         to perform a mode-by-mode cross check (unless `nmodes=None`).
        If the mode-by-mode cross check is a success, the solution is stored in the underlying `BGSystem` of the class.
        Otherwise, the `GEFSolver` tries to self correct using the gauge field spectrum. This is attempted for `mbm_attempts` or until successful.

        Parameters
        ----------
        ntr : int
            initial truncation number `GEFSolver.ntr`
        tend : float
            initial target time for `GEFSolver.tend`
        nmodes : int or None
            the number of modes computed by `ModeSolver`. If None, no cross-check is performed.
        mbm_attempts : int
            number of mode-by-mode self correction attempts
        resume_mbm : bool
            if `True` use `ModeSolver.update_spectrum` in case successive mode-by-mode comparisons are needed.
        err_tol : float
            passed to `mbm_crosscheck`.
        err_thr : float
            passed to `mbm_crosscheck`.
        binning : int
            passed to `mbm_crosscheck`.
        integrator : str
            integrator for `mbm_crosscheck` ('simpson' is advised)
        print_stats : bool
            if `True`, a summary report is printed for the returned solution.
        solver_kwargs
            the `settings` of `GEFSolver` (see `GEFSolver.settings`)

        Returns
        -------
        vals : BGSystem
            contains the background evolution obtained from `GEFSolver`
        spec : GaugeSpec or None
            the result of `ModeSolver.compute_spectrum`
        sol
            the full result of `GEFSolver.compute_GEF_solution`
        

        Raises
        ------
        RuntimeError
            if no successful solution was obtained.
        """
        solver = self.GEFSolver(self.initial_data)

        #Configuring GEFSolver
        solver.ntr=ntr
        solver.tend=tend
        solver.update_settings(**solver_kwargs)

        integrator_kwargs = {"integrator":integrator, "epsabs":solver.settings["atol"], "epsrel":solver.settings["rtol"]}

        done=False
        vals = BGSystem.from_system(self.initial_data, copy=True)
        vals.units = False
        sol = None
        spec = None
        t_reinit = 0.
        attempt=0

        print()

        while not(done) and attempt<mbm_attempts:
            attempt +=1
            #This can be taken care of internally. The GEF should not need to get sol objects...
            sol_new = solver.compute_GEF_solution()
            sol = self._update_sol(sol, sol_new)
            solver.parse_arr_to_sys(sol.t, sol.y, vals)

            if nmodes is not None:
                print("Using last successful GEF solution to compute gauge-field mode functions.")
                MbM = self.ModeSolver(vals)

                rtol = solver.settings["rtol"]

                if resume_mbm and attempt > 1:
                    spec = MbM.update_spectrum(spec, t_reinit, rtol=rtol)
                else:
                    spec = MbM.compute_spectrum(nmodes, rtol=rtol)
                print("Performing mode-by-mode comparison with GEF results.")

                agreement, reinit_spec = self.mbm_crosscheck(spec, vals, err_tol= err_tol, err_thr=err_thr, binning=binning,
                                                             **integrator_kwargs)

                if agreement:
                    print("The mode-by-mode comparison indicates a convergent GEF run.\n")
                    done=True
                
                else:
                    t_reinit = reinit_spec["t"]
                    print(f"Attempting to solve GEF using self-correction starting from t={reinit_spec['t']:.1f}, N={reinit_spec['N']:.1f}.\n")

                    solver.set_initial_conditions_to_MbM(sol, reinit_spec)
                
            else:
                done=True
        
        if done:
            if print_stats:
                self._print_summary(sol)
            if sol.success:
                vals.units = True
                return vals, spec, sol
            else:
                print("The run terminated on with an error, check output for details.")
        else:
            raise RuntimeError(f"GEF did not complete after {attempt} attempts.")
    
    @staticmethod
    def _print_summary(sol):
        print("GEF run completed with the following statistics")
        for attr in sol.keys():
            if attr not in ["y", "t", "y_events", "t_events", "sol", "events"]:
                print(rf" - {attr} : {getattr(sol, attr)}")
        events = sol.events
        if np.array([(len(event)==0) for event in events.values()]).all():
            print("No events occurred during the final run.")
        else:
            print("The following events occurred during the final run:")
            for event, time in events.items():
                if len(time) > 0:
                    if len(time) > 1:
                        tstr = [f"{t:.1f}" for t in time]
                    else:
                        tstr = f"{time[0]:.1f}"
                    print(f"  - {event} at t={tstr}")
        return
        
    @staticmethod
    def _update_sol(sol_old, sol_new):
        """
        Update an old GEF solution with a new one, overwriting the overlap.
        """
        if sol_old is None:
            return sol_new
        else:
            sol = sol_old
            ind_overlap = np.searchsorted(sol_old.t, sol_new.t[0], "left")
            sol.t = np.concatenate([sol_old.t[:ind_overlap], sol_new.t])

            if sol_old.y.shape[0] < sol_new.y.shape[0]:
                #if ntr increased from one solution to the next, fill up sol_old with zeros to match sol_new
                fillshape = (sol_new.y.shape[0] - sol_old.y.shape[0], sol_old.y.shape[1])
                yfill = np.zeros( fillshape )
                sol_old.y = np.concatenate([sol_old.y, yfill], axis=0)

            sol.y = np.concatenate([sol_old.y[:,:ind_overlap], sol_new.y], axis=1)
            sol.events.update(sol_new.events)
            for attr in ["nfev", "njev", "nlu"]:
                setattr(sol, attr, getattr(sol_old, attr) + getattr(sol_new, attr))
            for attr in ["message", "success", "status"]:
                setattr(sol, attr, getattr(sol_new, attr))
            return sol
        
    def mbm_crosscheck(self, spec, vals:BGSystem, err_tol:float, err_thr:float, binning:int, **integrator_kwargs):
        """
        Estimate the error of a GEF solution using `.mbm.GaugeSpec.estimate_GEF_error`.

        If either the RMS error or the final error exceeds `err_tol`, the solution is rejected.
        
        Parameters
        ----------
        vals : BGSystem
            contains the GEF solution.
        err_tol : float
            the tolerance on the RMS and final error.
        err_thr : float
            passed to `estimate_GEF_error`.
        binning : int
            passed to `estimate_GEF_error`.
        integratorkwargs:
            passed to kwargs of `estimate_GEF_error`.

        Returns
        -------
        agreement : bool
            indicates if the solution is accepted or rejected.
        reinit_slice : SpecSlice
            the spectrum with which the GEF solver is re-initialized.
        """
        GFs = self.GEFSolver.known_variables["gauge"]
        agreement=True

        for GF in GFs:
            ref = [a.name for a in GF.zeros]
            cut = GF.cutoff.name

            errs, terr, _ = spec.estimate_GEF_error(vals, references=ref, cutoff=cut,
                                                     err_thr=err_thr, binning=binning, **integrator_kwargs)

            reinit_inds = []
            
            for err in errs:
                rmserr = np.sqrt(np.sum(err**2)/len(err))
                if max(err[-1], rmserr) > err_tol:
                    agreement=False
                    #find where the error is above 5%, take the earliest occurrence, reduce by 1
                    err_ind = np.where(err > err_thr)[0][0]-1             
                else:
                    err_ind = len(terr)-1
                reinit_inds.append( err_ind )

            t0 = terr[min(reinit_inds)]

            ind = np.searchsorted(spec["t"],t0, "left")

            reinit_slice = spec.tslice(ind)

        return agreement, reinit_slice

    def load_GEFdata(self, path : str):
        """
        Load data and return a BGSystem with the results.

        Note, data is always loaded assuming numerical units.

        Parameters
        ----------
        path : None or str
            Path to load data from

        Raises
        ------
        FileNotFoundError
            if no file is found at `path`.
        AttributeError
            if the file contains a column labeled by a key which does not match any GEF-value name.
        """
        newsys = BGSystem.from_system(self.initial_data, copy=True)
        
        newsys.load_variables(path)

        return newsys

def compile_model(modelname:str, settings:dict={}):
    """
    Define a custom subclass of BaseGEF adapted to a new GEF model.

    Parameters
    ----------
    modelname : str
        The name of the GEF model or a full dotted import path (e.g., "path.to.module").
    settings : dict
        a dictionary of settings used by the model

    Returns
    -------
    CustomGEF
        a custom subclass of BaseGEF
        
    """
    model = load_model(modelname, settings)
    signature  = [inspect.Parameter("self", inspect.Parameter.POSITIONAL_ONLY)]
    for val in model.model_input:
        if issubclass(val, Func):
            annotation = Callable
        else:
            annotation = float
        signature.append(inspect.Parameter(val.name, inspect.Parameter.KEYWORD_ONLY, annotation=annotation))
    
    class CustomGEF(BaseGEF):
        GEFSolver = model.solver
        """The solver used to solve the GEF equations in `run`."""
        ModeSolver = model.MbM
        """The mode solver used for mode-by-mode cross checks."""
        _input_signature = model.model_input
        define_units = staticmethod(model.define_units)

    CustomGEF.__init__.__signature__ = inspect.Signature(signature)

    CustomGEF.__qualname__ = model.name
    CustomGEF.__module__ = __name__

    return CustomGEF

class ModelError(Exception):
    """
    Exception for errors when compiling a GEF model.
    """
    pass

generate_docs(docs_gef.DOCS)


"""
This module contains tools to compute physical observables from a GEF result.

Currently, the following tools are available:
1. Compute the tensor power spectrum using `geff.tools.pt`.
2. Compute and analyze the gravitational wave spectrum using `geff.tools.gw`.
"""
from .pt import PowSpecT
from .gw import omega_gw


"""
This module contains pre-defined GEF models.

Currently, the following models are implemented:
1. **'classic'**: pure axion inflation
2. **'SE noscale'**: fermionic axion inflation (without accounting for fermion $k$-dependence)
3. **'SE-kh'**: fermionic axion inflation (accounting for fermion $k$-dependence using $k_h$)

Execute these models by calling `GEFF.gef.GEF` with its name and settings.
"""
import importlib
import os

def load_model(model : str, user_settings : dict={}):
    """
    Import and execute a module defining a GEF model.

    Parameters
    ----------
    model : str
        The name of the GEF model or a full dotted import path (e.g., "path.to.module").
    user_settings : dict
        A dictionary containing updated settings for the module.

    Returns
    -------
    ModuleType
        The configured module.
    """

    # if model is  astring, load  ./models/{name}.py
    current_dir = os.path.dirname(os.path.abspath(__file__))
    modelpath = os.path.join(current_dir, f"{model}.py")

    if os.path.exists(modelpath):
        spec = importlib.util.spec_from_file_location(model, modelpath)
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)

    else:
        # if model is not a string, try loading it directly as a module
        try:
            mod = model
        except ImportError as e:
            raise FileNotFoundError(
                f"No model file found at '{modelpath}' and failed to import '{model}'"
                ) from e
        
    if hasattr(mod, "settings") and isinstance(user_settings, dict):
        for key, item in user_settings.items():
            if key in mod.settings:
                mod.settings[key] = item
                print(f"Updating '{key}' to '{item}'.")
            else:
                print(f"Ignoring unknown model setting '{key}'.")
        mod.interpret_settings()

    return mod


r"""
This module facilitates computing boundary terms which appear when computing the evolution of gauge-field bilinears like $\langle {\bf E}^2 \rangle$
due to the time dependence of the UV-regulator scale $k_{\rm h}$.

All functions in this module return the same quantities, namely
$$E_\lambda(\xi, s) = \frac{1}{r^2}\left| (i r - i \lambda \xi - s) W_{-i \lambda \xi, \frac{1}{2} + s}(-2 i r) + W_{1-i \lambda \xi, 1/2 + s}(-2 i r) \right|^2 \, , $$
$$B_\lambda(\xi, s) = \left| W_{-i \lambda \xi, \frac{1}{2} + s}(-2 i r) \right|^2 \, , $$
$$G_\lambda(\xi, s) = \frac{1}{r}\left[\operatorname{Re}\left[W_{1-i \lambda \xi, 1/2 + s}(-2 i r) W_{i \lambda \xi, \frac{1}{2} + s}(2 i r)\right] -  s \left| W_{-i \lambda \xi, \frac{1}{2} + s}(-2 i r)\right|^2 \right]\, ,$$
with $r = |\xi| + \sqrt{\xi^2 + s^2 + s}$, the Whittaker-W function $W_{\kappa, \mu}(x)$,the instability parameter  $\xi$ , and $s= \sigma_{\rm E}/(2H)$ an effective electric conductivity.

The functions in this module return an array of shape (3,2), with the first index corresponding to $E$, $B$, $G$ and the second index to helicity $\lambda=\pm 1$.
"""
import numpy as np
from scipy.special import gamma
from mpmath import whitw, mp
from functools import lru_cache

#set accuracy of mpmath
mp.dps = 8

def boundary_pai(xi) -> np.ndarray:
    """
    Compute boundary terms in the absence of Schwinger pair production.
    """
    if abs(xi) >= 3:
        return boundary_approx_pai(xi)
    else:
        return boundary_exact(xi, 0)
    
def boundary_fai(xi, s) ->  np.ndarray:
    """
    Compute boundary terms including Schwinger pair production.
    """
    if abs(xi) >= 4:
        return boundary_approx_fai(xi, s)
    else:
        return boundary_exact(xi, s)



@lru_cache(maxsize=int(1e6))
def whittaker_w(xi, s):
    r = (abs(xi) + np.sqrt(xi**2 + s**2 + s))
    w = whitw(-xi*1j, 1/2 + s, -2j*r)
    w1 = whitw(1-xi*1j, 1/2 + s, -2j*r)
    return complex(w), complex(w1)
    

def boundary_exact(xi : float, s : float) -> np.ndarray:
    """
    Compute boundary terms using exact Whittaker functions.

    Parameters
    ----------
    xi : float
    s : float

    Returns
    -------
    W : NDArray
    """
    xi = float(xi)
    wp, w1p = whittaker_w(xi, s)
    wm, w1m = whittaker_w(-xi, s)

    w = np.array([wp, wm])
    w1 = np.array([w1p, w1m])

    lam = np.array([1, -1])

    expterm = np.exp(np.pi*xi*lam)

    r = (abs(xi) + np.sqrt(xi**2 + s**2 + s))
    
    Fterm = np.zeros((3, 2))

    Fterm[0,:] = expterm*abs((1j*r - 1j**lam*xi -s) * w + w1)**2/r**2
    Fterm[1,:] = expterm*abs(w)**2
    Fterm[2,:] = expterm*((w1*w.conjugate()).real - s*abs(w)**2)/r

    return Fterm

def boundary_approx_pai(xi : float) -> np.ndarray:
    r"""
    Use approximate formulas to compute boundary terms for $|\xi| > 3$ and $s=0$.

    The approximations are taken from the Appendix B in [2109.01651](https://arxiv.org/abs/2109.01651).

    Parameters
    ----------
    xi : float

    Returns
    -------
    W : NDArray
    """
    if (abs(xi) >= 3):
        Fterm = np.zeros((3, 2))
        sgnsort = int((1-np.sign(xi))/2)

        xi = abs(xi)
        g1 = gamma(2/3)**2
        g2 = gamma(1/3)**2
        t1 = (3/2)**(1/3)*g1/(np.pi*xi**(1/3))
        t2 = -np.sqrt(3)/(15*xi)
        t3 = (2/3)**(1/3)*g2/(100*np.pi*xi**(5/3))
        t4 = (3/2)**(1/3)*g1/(1575*np.pi*xi**(7/3))
        t5 = -27*np.sqrt(3)/(19250*xi**3)
        t6 = 359*(2/3)**(1/3)*g2/(866250*np.pi*xi**(11/3))
        t7 = 8209*(3/2)**(1/3)*g1/(13162500*np.pi*xi**(13/3))
        t8 = -690978*np.sqrt(3)/(1861234375*xi**5)
        t9 = 13943074*(2/3)**(1/3)*g2/(127566140625*np.pi*xi**(17/3))
        Fterm[0, sgnsort] = t1+t2+t3+t4+t5+t6+t7+t8+t9

        t1 = 1
        t2 = -9/(2**(10)*xi**2)
        t3 = 2059/(2**(21)*xi**4)
        t4 = -448157/(2**31*xi**6)
        Fterm[0, 1-sgnsort] = np.sqrt(2)*(t1 + t2 + t3 + t4)

        t1 = (2/3)**(1/3)*g2*xi**(1/3)/(np.pi)
        t2 = 2*np.sqrt(3)/(35*xi)
        t3 = -4*(2/3)**(1/3)*g2/(225*np.pi*xi**(5/3))
        t4 = 9*(3/2)**(1/3)*g1/(1225*np.pi*xi**(7/3))
        t5 = 132*np.sqrt(3)/(56875*xi**3)
        t6 = -9511*(2/3)**(1/3)*g2/(5457375*np.pi*xi**(11/3))
        t7 = 1448*(3/2)**(1/3)*g1/(1990625*np.pi*xi**(13/3))
        t8 = 1187163*np.sqrt(3)/(1323765625*xi**5)
        t9 = -22862986*(2/3)**(1/3)*g2/(28465171875*np.pi*xi**(17/3))
        Fterm[1, sgnsort] = t1+t2+t3+t4+t5+t6+t7+t8+t9

        t1 = 1
        t2 = 11/(2**(10)*xi**2)
        t3 = -2397/(2**(21)*xi**4)
        t4 = 508063/(2**31*xi**6)
        Fterm[1, 1-sgnsort] = 1/np.sqrt(2)*(t1 + t2 + t3 + t4)

        t1 = 1/np.sqrt(3)
        t2 = -(2/3)**(1/3)*g2/(10*np.pi*xi**(2/3))
        t3 = 3*(3/2)**(1/3)*g1/(35*np.pi*xi**(4/3))
        t4 = -np.sqrt(3)/(175*xi**2)
        t5 = -41*(2/3)**(1/3)*g2/(34650*np.pi*xi**(8/3))
        t6 = 10201*(3/2)**(1/3)*g1/(2388750*np.pi*xi**(10/3))
        t7 = -8787*np.sqrt(3)/(21896875*xi**4)
        t8 = -1927529*(2/3)**(1/3)*g2/(4638768750*np.pi*xi**(14/3))
        t9 = 585443081*(3/2)**(1/3)*g1/(393158390625*np.pi*xi**(16/3))
        t10 = -65977497*np.sqrt(3)/(495088343750*xi**6)
        Fterm[2, sgnsort] = t1+t2+t3+t4+t5+t6+t7+t8+t9+t10

        t1 = 1
        t2 = -67/(2**(10)*xi**2)
        t3 = 21543/(2**(21)*xi**4)
        t4 = -6003491/(2**31*xi**6)
        Fterm[2, 1-sgnsort] = -np.sqrt(2)/(32*xi)*(t1 + t2 + t3 + t4) 
        return Fterm
    else:
        raise ValueError("abs(xi) needs to be larger than three for this approximation.")

def boundary_approx_fai(xi :float, s : float) -> np.ndarray:
    r"""
    Use approximate formulas to compute boundary terms for  $|\xi| > 4$ and arbitrary $s$.

    The approximations are taken from the Appendix B in [2109.01651](https://arxiv.org/abs/2109.01651).

    Parameters
    ----------
    xi : float

    Returns
    -------
    W : NDArray
    """
    if (abs(xi) >= 4):
        Fterm = np.zeros((3, 2))
        sgnsort = int((1-np.sign(xi))/2)
        
        xi = abs(xi)
        r = xi + np.sqrt(xi**2 + s**2 + s)
        psi = 2*np.sqrt(xi**2 + s**2 + s)/r
        rpsi = (psi/r**2)**(1/3)
        spsi = 5*s/psi**(2/3)
        
        g1 = gamma(2/3)**2
        g2 = gamma(1/3)**2
        
        t1 = 3**(1/3)*g1/np.pi
        t2 = -2/(5*np.sqrt(3))*(1+spsi)
        t3 = g2/(3**(1/3)*25*np.pi)*(1+spsi)**2
        t4 = 3**(1/3)*4*g1/(1575)*(1-27*spsi)
        t5 = 4*np.sqrt(3)/(875)*(-27/11+2*spsi + spsi**2)
        Fterm[0, sgnsort] = (psi/r)**(1/3)*(t1 + t2*rpsi + t3*rpsi**2 + t4*rpsi**3 + t5*rpsi**4)

        t1 = 1
        t2 = s/(16*xi**2)*(3*xi-r)/r
        t3 = s**2/(4*xi*r)
        Fterm[0, 1-sgnsort] = 2*np.sqrt(xi/r)*(t1 + t2 + t3)

        t1 = g2/(3**(1/3)*np.pi)
        t2 = 4*np.sqrt(3)/35
        t3 = -16*g2/(3**(1/3)*225*np.pi)
        Fterm[1, sgnsort] = (r/psi)**(1/3)*(t1 + t2*rpsi**2 + t3*rpsi**3)

        Fterm[1, 1-sgnsort] = 0.5*(r/xi)**(1/2)

        t1 = 1/np.sqrt(3)
        t2 = -g2/(3**(1/3)*5*np.pi)*(1+spsi)
        t3 = 3**(1/3)*6*g1/(35*np.pi)
        t4 = -4*np.sqrt(3)/(175)*(1+spsi)
        Fterm[2, sgnsort] = t1 + t2*rpsi + t3*rpsi**2 + t4*rpsi**3

        Fterm[2, 1-sgnsort] = -((3*xi -r)/xi + 8*s)/(16*np.sqrt(xi*r))
        return Fterm
    else:
        raise ValueError("abs(xi) needs to be larger than four for this approximation.")




"""
Some general purpose utility functions.
"""
import numpy as np
from typing import Callable

class AuxTol:
    """
    A simple wrapper which passes absolute and relative tolerance parameters.

    Wraps a function `f(*args, rtol, atol)` to give `f(*args, self.rtol, self.atol)` 


    """
    atol:float = 1e-20
    """Absolute tolerance"""
    rtol:float = 1e-6
    """Relative tolerance"""

    def __init__(self, func:Callable):
        self.func = func

    def __call__(self, *args):
        return self.func(*args, rtol=self.rtol, atol=self.atol)

@AuxTol
@np.vectorize
def heaviside(x : float, y:float, rtol:float, atol:float) -> float:
    r"""
    A smoothed version of the heaviside function.

    Returns 1 if x > y and 0 if x < y.
    
    The smoothness of the  transition is regulated by `max(abs(x)*rtol, atol))`.

    The function is wrapped by `AuxTol`, and will be updated according to the tolerances of a `GEFSolver`.

    It is vectorized using `numpy.vectorize`
    
    Parameters
    ----------
    x, y : floats
        the arguments of the heaviside function
    rtol, atol : floats
        tolerance parameters
    """
    eps = max(abs(x)*rtol, atol)
    arg = np.clip((x-y)/eps, -1e2, 1e18)
    return 1/(1+np.exp(-arg))


import os
import numpy as np
import pandas as pd
from scipy.interpolate import CubicSpline, PchipInterpolator
from scipy.integrate import solve_ivp
from scipy.integrate import quad, simpson
from .bgtypes import Variable, Constant, Func, BGSystem
from .utility.mode  import mode_equation_classic, bd_classic
from typing import Tuple, Callable, ClassVar
from types import NoneType
from ._docs import generate_docs, docs_mbm
from tabulate import tabulate


class GaugeSpec(dict):
    def __init__(self, in_dict : dict):
        """
        Initialise the spectrum from a dictionary.

        Parameters
        ----------
        in_dict : dict
            dictionary with keys `'t'`, `'N'`, `'k'`, `'Ap'`, `'dAp'`, `'Am'`, `'dAm'`

        Raises
        ------
        KeyError
            if one of the necessary keys is missing.
        ValueError
            if `len(in_dict['t']) != len(in_dict['N'])` or if
            `in_dict['X']).shape != (len(in_dict['k']),len(in_dict['t']))` for `'X'` in `['Ap', 'dAp', 'Am', 'dAm']`.
        """
        for key in ["t", "N", "k", "Ap", "dAp", "Am", "dAm"]:
            if key not in in_dict.keys():
                raise KeyError(f"Missing key: {key}")
            
        if not(len(in_dict["t"]) == len(in_dict["N"])):
            raise ValueError("The length of 't' needs to match the length of 'N'")
        
        shape = (len(in_dict["k"]), len(in_dict["t"]))
        for key in ["Ap", "dAp", "Am", "dAm"]:
            if not( in_dict[key].shape == shape):
                raise ValueError(f"The shape of {key} needs to be {shape}") 
        
        super().__init__(in_dict)

    @classmethod
    def read_spec(cls, path : str):
        """
        Initialise the class from a file.

        Parameters
        ----------
        path : str
            path to the data 

        Returns
        -------
        spec : GaugeSpec
            the imported spectrum
        """
        input_df = pd.read_table(path, sep=",")
        dataAp = input_df.values

        x = np.arange(3,dataAp.shape[1], 4)
        
        t = np.asarray(dataAp[1:,1], dtype=float)
        N = np.asarray(dataAp[1:,2], dtype=float)
        logk = np.array([(complex(dataAp[0,y])).real for y in x])
        Ap = np.array([[complex(dataAp[i+1,y]) for i in range(len(N))] for y in x])
        dAp = np.array([[complex(dataAp[i+1,y+1]) for i in range(len(N))] for y in x])
        Am = np.array([[complex(dataAp[i+1,y+2]) for i in range(len(N))] for y in x])
        dAm = np.array([[complex(dataAp[i+1,y+3]) for i in range(len(N))] for y in x])

        k = 10**logk

        spec = cls({"t":t, "N":N, "k":k,
                        "Ap":Ap, "dAp":dAp, "Am":Am, "dAm":dAm})
        
        return spec
    
    def save_spec(self, path : str):
        """
        Store the spectrum in a file.

        Parameters
        ----------
        path : str
            path to the data 
        """
        thinning = 1
        N = np.array([np.nan]+list(self["N"][-1::-thinning][::-1]))
        t = np.array([np.nan]+list(self["t"][-1::-thinning][::-1]))

        dic = {"t":t, "N":N}

        for j, k in enumerate(self["k"]):
            spec_slice = self.kslice(j)
            logk = np.log10(spec_slice["k"])
            for key in ["Ap", "dAp", "Am", "dAm"]:
                dictmp = {key + "_" + str(j) :np.array([logk] + 
                                                       list(spec_slice[key][-1::-thinning][::-1]))}
                dic.update(dictmp)
    
        dir_name = os.getcwd()
    
        path = os.path.join(dir_name, path)
    
        output_df = pd.DataFrame(dic)  
        output_df.to_csv(path)
        
        return

    def get_dim(self) -> dict:
        """
        Get the number of time coordinates and momenta stored in the spectrum.

        Returns
        -------
        dict
            a dictionary encoding the spectrum's shape
        """
        return {"kdim":len(self["k"]), "tdim":len(self["t"])}
    
    def tslice(self, ind : int) -> dict:
        """
        Evaluate the spectrum at time `self['t'][ind]`.

        Parameters
        ----------
        ind : int
            the temporal index

        Returns
        -------
        spec_slice : SpecSlice
            the spectrum at time `self['t'][ind]`
        """

        spec_slice = {}
        for key in self.keys():
            if key in ["N", "t", "cut"]:
                spec_slice[key] = self[key][ind]
            elif key=="k":
                spec_slice[key] = self[key]
            else:
                spec_slice[key] = self[key][:,ind]
        return SpecSlice(spec_slice)
    
    def kslice(self, ind : int) -> dict:
        """
        Obtain the time evolution for the momentum `self['k'][ind]`.

        Parameters
        ----------
        ind : int
            the momentum index.

        Returns
        -------
        spec_slice : dict
            a dictionary with keys like `self`  
        """

        spec_slice = {}
        for key in self.keys():
            if key in ["N", "t", "cut"]:
                spec_slice[key] = self[key]
            elif key=="k":
                spec_slice[key] = self[key][ind]
            else:
                spec_slice[key] = self[key][ind,:]
        return spec_slice
    
    def integrate(self, BG : BGSystem, n : int=0, cutoff="kh", **IntegratorKwargs) -> np.ndarray:
        r"""
        Compute the three integrals $\mathcal{F}_\mathcal{X}^{(n)}(t)$ for $\mathcal{X} = \mathcal{E}, \mathcal{B},\mathcal{G}$ for fixed $n$ and each time $t$ in the spectrum.

        If the time coordinates stored in `BG` do not match those stored in the spectrum, $k_{\rm UV}(t)$ is evaluated using interpolation.

        Parameters
        ----------
        BG : BGSystem
            a system containing the UV cut-off, $k_{\rm UV}(t)$
        n : int
            the integer $n$
        cutoff : str
            the name under which the UV-cutoff is stored in `BG`
        **IntegratorKwargs :  kwargs
            passed to `SpecSlice.integrate_slice`
        

        Returns
        -------
        FMbM : NDArray
            $\mathcal{F}_\mathcal{E}^{(n)}(t)$, $\mathcal{F}_\mathcal{B}^{(n)}(t)$, $\mathcal{F}_\mathcal{B}^{(n)}(t)$ stored in a shape (N, 3, 2).
            The first index corresponds to time $t$, the second index to $\mathcal{X}$, the third index is the integral result (at 0) and its error (at 1).
        """
        
        self._add_cutoff(BG, cutoff)

        tdim = self.get_dim()["tdim"]

        FMbM = np.zeros((tdim, 3,2))
        for i in range(tdim):
            spec_slice = self.tslice(i)
            FMbM[i,:] = spec_slice.integrate_slice(n=n, **IntegratorKwargs)
        
        return FMbM
    
    def estimate_GEF_error(self, BG : BGSystem, references : list[str]=["E", "B", "G"], cutoff : str="kh",
                                    err_thr : float=0.025, binning : int|NoneType=5, verbose : bool=True,
                                    **IntegratorKwargs) -> Tuple[list, np.ndarray, list]:
        

        og_errs = self._estimate_error(BG, references, cutoff, **IntegratorKwargs)
        terr = self["t"]

        if binning is not None:
            errs, terr = self._bin_error(og_errs, binning)
        else:
            errs = og_errs
        
        errs, terr = self._process_error(errs, terr, err_thr)

        if verbose:
            self._error_summary(errs, terr, references)

        return errs, terr, og_errs
    
    def merge_spectra(self, spec : 'GaugeSpec'):
        """
        Combine two spectra with the same momenta $k$ but unequal times $t$.

        Parameters
        ----------
        spec : GaugeSpec
            the second spectrum

        Raises
        ------
        AssertionError
            if the momenta $k$ do not match up.
        
        """
        assert (spec["k"] == self["k"]).all()

        ind = np.searchsorted(self["t"], spec["t"][0], "left")

        if "cut" in self.keys():
            self.pop("cut")

        for key in self.keys():
            if key in ["t", "N"]:
                self[key] = np.concatenate([self[key][:ind], spec[key]])
            else:
                if key != "k":
                    self[key] = np.concatenate([self[key][:,:ind], spec[key]], axis=1)
        return
    
    def add_momenta(self, spec : 'GaugeSpec'):
        """
        Combine two spectra with the same times $t$ but unequal momenta $k$.

        Parameters
        ----------
        spec : GaugeSpec
            the second spectrum

        Raises
        ------
        AssertionError
            if the times $t$ do not match up.
        
        """

        assert (np.round(spec["t"],1) == np.round(self["t"],1)).all()

        newks = []
        mds = ["Ap", "dAp", "Am", "dAm"]
        newmodes = dict( zip(mds, [ [] for i in mds]) )

        for i, k in enumerate(self["k"]):
            mask = np.where(k > spec["k"])[0]
            if len(mask) != 0:
                newks.append(spec["k"][mask])    
                spec["k"] = np.delete(spec["k"], mask)
                for md in mds:
                    newmodes[md].append(spec[md][mask,:])
                    spec[md] = np.delete(spec[md], mask, axis=0)
            newks.append(np.array([k]))
            for md in mds:
                newmodes[md].append(np.array([self[md][i,:]]))
        
        mask = np.where(spec["k"] > k)[0]
        if len(mask) != 0:
            newks.append(spec["k"][mask])    
            spec["k"] = np.delete(spec["k"], mask)
            for md in mds:
                newmodes[md].append(spec[md][mask,:])
                spec[md] = np.delete(spec[md], mask, axis=0)

        self["k"] = np.concatenate(newks)
        for md in mds:
            self[md] = np.concatenate(newmodes[md], axis=0)

        return
    
    def remove_momenta(self, ind):
        """
        Remove the spectrum at momentum `self["k"][ind]`.

        Parameters
        ----------
        ind : int
            the index at which to remove the spectrum entry
        """
        self["k"] = np.delete(self["k"], ind)
        for md in ["Ap", "dAp", "Am", "dAm"]:
            self[md] = np.delete(self[md], ind, axis=0)
        return
    
    def _check_overlap(self, t):
        mask = np.isin(t, self["t"], assume_unique=True)
        if len(t[mask]) != len(self["t"]):
            #this should be a warning
            print("The times in the current GaugeSpec instance are " \
            "not a subset of the times in the BGSystem. Reverting to interpolation.")

            return False, None
        else:
            return True, mask
    
    def _add_cutoff(self, BG : BGSystem, cutoff="kh"):
        units = BG.units
        BG.units = (False)

        scale = getattr(BG, cutoff)

        bl, mask = self._check_overlap(BG.t)

        if bl:
            self["cut"] = scale[mask]
        else:
            self["cut"] = CubicSpline(BG.t, scale)(self["t"])
        
        BG.units = (units)

        return self["cut"]
    
    def _get_reference(self, BG : BGSystem, references=["E", "B", "G"], cutoff="kh"): 
        units = BG.units
        BG.units = (False)

        scale = getattr(BG, cutoff)

        bl, mask = self._check_overlap(BG.t)

        Fref = []
        for val in references: 
            val_arr = (getattr(BG, val)*(np.exp(BG.N)/scale)**4)
            if bl:
                Fref.append( val_arr[mask] )
            else:
                Fref.append( CubicSpline(BG.t, val_arr)(self["t"]) ) 

        BG.units = (units)

        return Fref
    
    
    def _estimate_error(self, BG : BGSystem, references : list[str]=["E", "B", "G"], cutoff : str="kh",
                         **IntegratorKwargs):
        FMbM = self.integrate(BG, n=0, cutoff=cutoff, **IntegratorKwargs)
        Fref = self._get_reference(BG, references, cutoff)

        errs = []

        for i, spl in enumerate(Fref):
            err =  np.minimum(1e3, abs( 1 - FMbM[:,i,0]/(spl+1e-20) )) #avoid divide by zero
            errs.append(np.where(np.isnan(err), 10.0, err))

        return errs
    
    def _process_error(self, errs, terr, err_thr):
        removals = []
        for err in errs:
            #remove the first few errors where the density of modes is low:
            removals.append(np.where(err < err_thr)[0][0])
        #ind = 0
        ind = max(removals)
        errs = [err[ind:] for err in errs]
        terr = terr[ind:]

        return errs, terr
    
    def _bin_error(self, errs, binning):
        terr = self["t"]

        bin_terr = terr[::-binning][::-1]
        tbins = bin_terr[1:] + (bin_terr[1:] - bin_terr[:-1])/2
        count, _  = np.histogram(terr, bins=tbins)

        bin_errs = []
        for err in errs:
            sum, _  = np.histogram(terr, bins=tbins, weights=err)
            bin_errs.append(sum/count)
        bin_terr = bin_terr[2:] 

        return bin_errs, bin_terr
    
    @staticmethod
    def _error_summary(bin_errs, bin_terr, references : list[str]=["E", "B", "G"]):
        print("The mode-by-mode comparison finds the following relative deviations from the GEF solution:\n")

        lst = []
        for err in bin_errs:
            rmserr = 100*np.sqrt(np.sum(err**2)/len(err))
            finerr = 100*err[-1]
            maxerr = 100*max(err)

            tmaxerr = bin_terr[np.argmax(err)]
            tfinerr = bin_terr[-1]

            lst.append([f"{maxerr:.1f}% at {tmaxerr:.1f}",
                        f"{finerr:.1f}% at {tfinerr:.1f}",
                        f"{rmserr:.1f}%"])

        print(tabulate(lst, headers=["max", "final", "RMS"], showindex=references, tablefmt="simple")+"\n")
        return

   
# continue from here next time
class SpecSlice(dict):    

    def __init__(self, in_dict):
        super().__init__(in_dict)

    def integrate_slice(self, n : int=0, integrator="simpson", modethr : int=100,
                         epsabs : float=1e-20, epsrel : float=1e-4, interpolator=PchipInterpolator
                         )-> Tuple[np.ndarray, np.ndarray]:
        x = (n+4)*np.log(self["k"]/self["cut"])

        helicities = ["p", "m"]
        integs = np.zeros((3, 2, len(x)))
        for i, lam in enumerate(helicities):
            sgn = np.sign(0.5-i)
            integs[0,i,:] = self._Espec(lam)
            integs[1,i,:] = self._Bspec(lam)
            integs[2,i,:] = sgn*self._Gspec(lam)

        res = np.zeros((3,2))
        
        for i in range(3):
            if integrator=="simpson":
                msk = np.where(x < 0)[0]
                if len(msk) < modethr: #cannot trust simpsons integration for too few modes.
                    return res
                x = x[msk]
                res[i,0] = (self._simpson_integrate( integs[i,0,msk] ,x) 
                                         + (-1)**n*self._simpson_integrate(integs[i,1,msk], x) )
                res[i,1] = 1e-6*res[i,0]

            elif integrator=="quad":
                resp = self._quad_integrate( integs[i,0,:], x, epsabs, epsrel, interpolator)
                resm = self._quad_integrate( (-1)**n*integs[i,1,:], x, epsabs, epsrel, interpolator)
                res[i,:] = resp +resm
            
        res = 1/(2*np.pi)**2*res/(n+4)

        res[:,1] = abs(res[:,1]/res[:,0])
        return res

    def _Espec(self, lam):
        return abs(self["dA"+lam])**2
    
    def _Bspec(self, lam):
        return abs(self["A"+lam])**2
    
    def _Gspec(self, lam):
        return (self["A"+lam].conjugate()*self["dA"+lam]).real
    
    def _simpson_integrate(self, integrand, x):
        integrand = integrand*np.exp(x)
        return simpson(integrand, x)

    def _quad_integrate(self, integrand, x, epsabs : float=1e-20, epsrel : float=1e-4, interpolator=PchipInterpolator):
        msk = np.where(abs(integrand) > 1e-1*max(epsrel*max(abs(integrand)), epsabs))[0]
        if len(msk) > 0:
            spl = interpolator(x, np.arcsinh(integrand))
            def f(x): return np.sinh(spl(x))*np.exp(x)
            val, err = quad(f, x[msk][0], 0., epsabs=epsabs, epsrel=epsrel)
            return np.array([val, err])
        else:
            return np.nan*np.ones((2))

class BaseModeSolver:    
    _ode_dict : dict = {"a": "a", "H":"H", "xi":"xi"}
    _bd_dict = {}

    cutoff : ClassVar[str] = "kh"
    r"""The name of $k_{\rm UV}(t)$ in the GEF solution."""
    
    atol : ClassVar[float] = 1e-3
    """The default absolute tolerance used in `scipy.integrate.solve_ivp`"""

    necessary_keys : ClassVar[set] = (["t", "N", "a", "H", "xi"] + [cutoff])
    """The class expects these attributes in the `.bgtypes.BGSystem` passed on initialisation."""

    mode_equation = staticmethod(mode_equation_classic)
    initialise_in_bd = staticmethod(bd_classic)

    def __init__(self, sys : BGSystem):
        """
        Import the evolution of the background dynamics to configure the solver.

        All values in the BGSystem are treated in numerical units.

        Parameters
        ----------
        sys : BGSystem
            describes the background evolution
        
        Raises
        ------
        KeyError:
            if `sys` is missing a `Val` or `Func` object from `necessary_keys`
        ValueError:
            if the keys in `necessary_keys` are not `Val` or `Func` objects.
        """
        og_units = sys.units
        sys.units = False

        #Check that all necessary keys are there:
        for key in self.necessary_keys:
            check = True
            check = ( (key in sys.variable_names())
                        or
                        (key in sys.constant_names())
                        or 
                        (key in sys.function_names())
                    )
            if not(check):
                raise KeyError(f"'sys' needs to own an attribute called '{key}'.")

        #store the relevant background evolution parameters
        self._t = sys.t.value
        self._N = sys.N.value
        kh = getattr(sys, self.cutoff)
        a = np.exp(self._N)

        self._khf = CubicSpline(self._t, kh)

        self._kwargs_ode = {kwarg:None for kwarg in self._ode_dict.values()}
        self._kwargs_bd = {kwarg:None for kwarg in self._bd_dict.values()}

        #import the values for the mode equation and interpolate
        for key in self.necessary_keys: 
            obj = getattr(sys, key)
            if isinstance(obj, Variable):
                value = getattr(sys, key).value
                func = CubicSpline(self._t, value)
            elif isinstance(obj, Func):
                arg_vals = []
                for arg in obj.args:
                    arg_vals.append(getattr(sys, arg.name))
                value = obj(*arg_vals)
                func =  CubicSpline(self._t, value)
            elif isinstance(obj, Constant):
                value = getattr(sys, key).value
                func =  CubicSpline(self._t, value*np.ones_like(self._t))
            else:
                raise ValueError(f"'{key}' should refer to either a 'Val' or 'Func' subclass.")
                
            if key in self._ode_dict.keys():
                kwarg = self._ode_dict[key]
                self._kwargs_ode[kwarg] = func
            if key in self._bd_dict.values():
                kwarg = self._bd_dict[key]
                self._kwargs_bd[kwarg] = func
        
        #compute the evolution of conformal time for the phases
        af = CubicSpline(self._t, a)
        def deta(t, y): return 1/af(t)
        
        soleta = solve_ivp(deta, [min(self._t), max(self._t)], np.array([0]), t_eval=self._t)

        self.__eta = soleta.y[0,:]

        #find lowest t value corresponding to kh(t) = 10^4 kh(0)
        self.__tmin = self._t[np.searchsorted(kh, 10**4*kh[0], "right")]

        sys.units = og_units
        
        return
    
    def compute_spectrum(self, nvals : int, t_interval : tuple|NoneType=None, **SolverKwargs) -> GaugeSpec:
        r"""
        Evolve a gauge-field spectrum from Bunch-Davies initial conditions.

        Evolve the mode functions $A_\lambda(t,k)$ and its derivative in time for $n$ modes between $k_{\rm UV}(t_{\rm min})$ and $k_{\rm UV}(t_{\rm max})$.
        The $n$ evolved modes are more densly spaced when $\log k_{\rm UV}(t)$ increases more slowly to ensure a higher density of modes
        crossing the horizon when backreaction effects are relevant.

        Parameters
        ----------
        nvals : int
           The number of modes $n$
        t_interval : tuple or None
            set $t_{\rm min}$ and $t_{\rm max}$. If None, $t_{\rm min}$ is given by $10^4 k_{\rm UV}(t_{min}) = k_{\rm UV}(0)$ and $t_{\rm max} = \max t$.
        **SolverKwargs : kwargs
            tolerance parameters`atol` and `rtol` passed to `solve_ivp` (default: `atol=self.atol`, `rtol=1e-5`)

        Returns
        -------
        spec : GaugeSpec 
            the gauge-field spectrum
        """

        if t_interval is None:
            t_interval = (self.__tmin, max(self._t))
        ks, tstart = self._find_tinit_bd(self._create_k_array(nvals, t_interval), mode="k")

        modes = np.array([self._evolve_from_bd(k, tstart[i], **SolverKwargs)
                  for i, k in enumerate(ks)])
        
        spec = GaugeSpec({"t":self._t, "N":self._N, "k":ks,
                    "Ap":modes[:,0,:], "dAp":modes[:,1,:], "Am":modes[:,2,:], "dAm":modes[:,3,:]})

        return spec
    
    def update_spectrum(self, spec : GaugeSpec, tstart : float, **SolverKwargs) -> GaugeSpec:
        r"""
        Update an existing gauge-field spectrum starting from $t_{\rm start}$

        Starting from the modes stored in `GaugeSpec`, the function re-evaluates the evolution starting from $t_{\rm start}$.
        Additional gauge-field modes are evolved starting from Bunch&ndash;Davies to account for new modes crossing the horizon
          at times beyond the original range covered by the input spectrum.

        Parameters
        ----------
        spec : GaugeSpec
           the spectrum which is to be updated
        tstart : float
            the starting time $t_{\rm start}$
        **SolverKwargs : kwargs
            as in `compute_spectrum`

        Returns
        -------
        spec : GaugeSpec 
            the updated gauge-field spectrum
        """
        
        indstart = np.searchsorted(self._t, tstart, "left")
        teval = self._t[indstart:]
        Neval = self._N[indstart:]
        
        tend = teval[-1]
        indstart = np.searchsorted(spec["t"], teval[0], "left")
        startspec = spec.tslice(indstart)
        tstart = startspec["t"]

        #keep mode-evolution from old spectrum for modes with k < 10*kh(tstart)
        old = np.where(spec["k"] < 10*self._khf(tstart))
        new = np.where(spec["k"] >= 10*self._khf(tstart))

        #Remove modes which need to be renewed from old spectrum:
        spec.remove_momenta(new)

        #add new modes, adjusting for longer time-span:
        #rethink this! It seems like overkill
        n_newmodes = int(len(new)*max((teval[-1] - teval[0])/(max(spec["t"]) - tstart), 1))

        #Update evolution of modes in spec:
        kold, tvac = self._find_tinit_bd(spec["k"][old], mode="k")

        updatespec={"t":teval, "N":Neval, "k":kold}

        modes = []
        for i, k in enumerate(kold):
            if tvac[i] > teval[0]:
                modes.append( self._evolve_from_bd(k, tvac[i], **SolverKwargs) )
            else:
                yini = np.array(
                            [startspec["Ap"][i].real, startspec["dAp"][i].real,
                            startspec["Ap"][i].imag, startspec["dAp"][i].imag,
                            startspec["Am"][i].real, startspec["dAm"][i].real,
                            startspec["Am"][i].imag, startspec["dAm"][i].imag]
                            )
                
                modes.append( self._evolve_mode(tstart, yini, k, teval, **SolverKwargs) ) 
        
        modes = np.array(modes)

        updatespec.update({"Ap":modes[:,0,:], "dAp":modes[:,1,:], "Am":modes[:,2,:], "dAm":modes[:,3,:]})

        spec.merge_spectra(GaugeSpec(updatespec))

        if n_newmodes > 0:
            newspec = self.compute_spectrum(n_newmodes, t_interval=(tstart, tend))
            #Add new modes
            spec.add_momenta(newspec)
        
        return spec
    
        """
        def EvolveSpectrum(self, spec : GaugeSpec, Nstart, Nstep : float=0.1, atol : float|None=None, rtol : float=1e-5) -> GaugeSpec:
            Neval = np.arange(Nstart, max(self.__N), Nstep)
            teval = CubicSpline(self.__N, self.__t)(Neval)

            indstart = np.where(spec["N"]<Nstart)[0][-1]
            startspec = spec.tslice(indstart)

            klen = len(spec["k"])

            vecode = np.vectorize(lambda t, y, k: self.mode_equation(t, y, k, **self._ode_kwargs),
                                    excluded={0, "t"},
                                signature="(8,n),(n)->(8,n)",
                                )
            def ode(t, y):
                #(k,8) to reshape correctly
                #transposing to match signature of vecode
                y = y.reshape(klen,8).T
                #transposing result s.t. dydt.shape=(k,8)
                dydt  = vecode(t, y, spec["k"]).T
                #reshape is correct again
                return dydt.reshape(8*klen)
            
        spec.merge_spectra(GaugeSpec(newspec))

            return spec
        """
    

    def _evolve_from_bd(self, k : float, tstart : float,
                    atol : float|None=None, rtol : float=1e-5) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """
        Evolve a mode with momentum $k$ starting from Bunch-Davies initial conditions.

        Parameters
        ----------
        k : float
           the momentum $k$
        tstart : float
            the corresponding initialisation time, $t_{\rm init}$
        atol : float or None
            `atol` used by `solve_ivp` if None, use `self.atol`
        rtol : float
            `rtol` used by `solve_ivp`

        Returns
        -------
        yp : NDArray
            the positive helicity mode
        dyp : NDArray
            the derivative of the positive helicity mode
        ym : NDArray
            the positive helicity mode
        dym : NDArray
            the derivative of the negative helicity mode
        """

        #Initial conditions for y and dydt for both helicities (rescaled appropriately)
        yini = self.initialise_in_bd(tstart, k, **self._kwargs_bd)

        teval = self._t

        istart = np.searchsorted(teval, tstart, "right")

        yp, dyp, ym, dym = self._evolve_mode(tstart, yini, k, teval[istart:], atol, rtol)

        #conformal time needed for relative phases
        eta = self.__eta
        
        #the mode was in vacuum before tstart
        yvac = np.array([self.initialise_in_bd(t, k, **self._kwargs_bd) for t in teval[:istart]]).T 
        phasevac = (np.exp(-1j*k*eta[:istart]))
        vac = yvac * phasevac

        #Create array of mode evolution stringing together vacuum and non-vacuum time evolutions to get evolution from t0 to tend
        yp = np.concatenate([(vac[0,:] + 1j*vac[2,:]), yp*np.exp(-1j*k*eta[istart])])
        dyp = np.concatenate([(vac[1,:] + 1j*vac[3,:]), dyp*np.exp(-1j*k*eta[istart])])
        ym = np.concatenate([(vac[4,:] + 1j*vac[6,:]), ym*np.exp(-1j*k*eta[istart])])
        dym = np.concatenate([(vac[5,:] + 1j*vac[7,:]), dym*np.exp(-1j*k*eta[istart])])

        return yp, dyp, ym, dym
    
    def _evolve_mode(self, tini : float, yini : np.ndarray, k : float, teval : np.ndarray,
                    atol : float|None=None, rtol : float=1e-5):
        """
        Evolve a mode of momentum $k$ from $t_{\rm ini}$.

        Parameters
        ----------
        tini : float
           the initial time $t_{\rm ini}$
        yini : NDArray
            (8,) array containing the initial data
        k : float
            the momentum $k$
        teval : NDArray
            the times at which the returned solution is evaluated
        atol : float or None
            `atol` used by `solve_ivp` if None, use `self.atol`
        rtol : float
            `rtol` used by `solve_ivp`

        Returns
        -------
        yp : NDArray
            the positive helicity mode
        dyp : NDArray
            the derivative of the positive helicity mode
        ym : NDArray
            the positive helicity mode
        dym : NDArray
            the derivative of the negative helicity mode
        """
        #Define ODE
        def ode(t, y): return self.mode_equation(t, y, k, **self._kwargs_ode)

        if atol is None:
            atol = self.atol
        
        #Solve differential equation from tstart to tmax
        sol = solve_ivp(ode, [tini, max(teval)], yini, t_eval=teval, method="RK45", atol=atol, rtol=rtol)

        yp = (sol.y[0,:] + 1j*sol.y[2,:])
        dyp = (sol.y[1,:] + 1j*sol.y[3,:])
        ym = (sol.y[4,:] + 1j*sol.y[6,:])
        dym = (sol.y[5,:] + 1j*sol.y[7,:])

        return yp, dyp, ym, dym 
        

    def _create_k_array(self, nvals : int, t_interval : tuple) -> np.ndarray:
        r"""
        Create an array of $n$ momenta

        The $n$ modes are generated between $k_{\rm UV}(t_{\rm min})$ and $k_{\rm UV}(t_{\rm max})$.
        First, $m$ of modes $k = k_{\rm UV}$ with t evenly spaced between $t_{\rm min}$ and $t_{\rm max}$ are generated.
        As $k_{\rm UV}$ is monotonic but not strictly monotonic, $m\leq n$. To fill up to $n$ modes, $n-m$ additional modes are
        created between the existing times $(t_{i},t_{i+1})$ moving backwards from $t_{\rm max}$ to favour larger momenta. 

        Parameters
        ----------
        nvals : int
            The size of the output array
        t_interval : tuple
            sets $t_{\rm min}$ and $t_{\rm max}$

        Returns
        -------
        NDArray
            an array of momenta with size nvals
        """

        #create an array of values log(10*kh(t))
        logks = np.round( np.log(10*self._khf(np.linspace(t_interval[0], t_interval[1], nvals))), 3)
        #filter out all values that are repeating (kh is not strictly monotonous)
        logks = np.unique((logks))

        #fill up the array of ks values with additional elements between gaps, favouring larger k
        while len(logks) < nvals:
            num_newk = nvals - len(logks)
            if num_newk >= len(logks):
                newvals = (logks[1:] + logks[:-1])/2
            else:
                newvals = (logks[-num_newk:] + logks[-num_newk-1:-1])/2
            logks = np.sort(np.concatenate([logks, newvals]))
        return np.exp(logks)
    
    def _find_tinit_bd(self, init : np.ndarray, mode : str="k") -> Tuple[np.ndarray, np.ndarray]:
        """
        Determines the pair of $k$ and $t$ satisfying $k = 10^(5/2)k_h(t)$.

        Depending on `mode`, `init` may be a time coordinate (`mode='t'`), $e$-folds (`mode='N'`) or momentum (`mode='k'`).

        Parameters
        ----------
        init : array
            the input array (t, N, or k)
        mode : str
            indicate the type of `init`

        Returns
        -------
        ks : NDarray
            an array of momenta
        tstart : NDarray
            an array of times

        Raises
        ------
        KeyError
            if `mode` is not 't, 'k' or 'N'
        """

        if mode=="t":
            tstart = init
            ks = 10**(5/2)*self._khf(tstart)

        elif mode=="k":
            ks = init
            
            tstart = []
            for k in ks:
                ttmp  = self._t[np.searchsorted(10**(5/2)*self._khf(self._t), k, "right")]
                tstart.append(ttmp)
            tstart = np.array(tstart)

        elif mode=="N":
            tstart = CubicSpline(self._N, self._t)(init)
            ks = 10**(5/2)*self._khf(tstart)

        else:
            raise KeyError("'mode' must be 't', 'k' or 'N'")

        return ks, tstart
        
    
def ModeSolver(new_mode_eq : Callable, ode_keys : list[str], new_bd_init : Callable, init_keys : list[str], new_cutoff : str="kh", default_atol : float=1e-3):
    class CustomModeSolver(BaseModeSolver):
        """
        A subclass of `BaseModeSolver` new mode equation and initial condition adapted to a new GEF model

        It Inherits all methods from `BaseModeSolver` but changes the attributes:
        - `BaseModeSolver.mode_equation`
        - `BaseModeSolver.initialise_in_bd`
        - `BaseModeSolver.necessary_keys`
        - `BaseModeSolver.cutoff`
        - `BaseModeSolver.atol`
            
        This entails that `compute_spectrum` will now evolve modes according to `initialise_in_bd` and `mode_equation`.
        """
        
        #Overwrite class attibutes of ModeByMode with new mode equations, boundary conditions and default tolerances.
        mode_equation = staticmethod(new_mode_eq)
        _ode_dict = {attr.name:kwarg for kwarg, attr in ode_keys.items()}

        initialise_in_bd = staticmethod(new_bd_init)
        _bd_dict = {attr.name:kwarg for kwarg, attr in init_keys.items()}
        cutoff = new_cutoff

        necessary_keys = list( {"t", "N", new_cutoff}.union(set(_ode_dict.keys())).union(set(_bd_dict.keys())) )

        atol=default_atol

    CustomModeSolver.__qualname__ = "ModeSolver"
    CustomModeSolver.__module__ = __name__
    return CustomModeSolver


#define longer method docs from docs_mbm:
generate_docs(docs_mbm.DOCS)







    
    
    
    
    
    
    
    
    


def generate_docs(docs, namespace=None):
    """
    Dynamically assign docstrings to functions, methods, or classes in a module.

    Args:
        docs (dict): A dictionary mapping names (e.g., "Class.method" or "function") to docstrings.
        namespace (dict, optional): The namespace (e.g., globals() or locals()) of the module where the objects are defined.
                                    If None, defaults to the globals() of the caller.
    """
    if namespace is None:
        # Use the globals() of the caller if no namespace is provided
        import inspect
        namespace = inspect.currentframe().f_back.f_globals

    for name, docstring in docs.items():
        if "module" in name:
            namespace["__doc__"] = docstring
        if "." in name:
            # Handle methods (e.g., "Class.method")
            classname, methodname = name.split(".")
            if classname in namespace:
                cls = namespace[classname]
                method = getattr(cls, methodname, None)
                if method is not None and hasattr(method, "__doc__"):
                    method.__doc__ = docstring
        else:
            # Handle functions or classes (e.g., "function" or "Class")
            obj = namespace.get(name, None)
            if obj is not None and hasattr(obj, "__doc__"):
                obj.__doc__ = docstring


DOCS = {
    "module":"""
    This module defines the `BaseGEF` class and its class factory `compile_model`.
    """,
    "BaseGEF":"""
    This class defines the basic properties of a GEF model.

    Each GEF model contains a `GEFSolver` and a `ModeSolver` that are used to solve GEF equations with the `run` method.

    Instances of a GEF model can also be used to load data using `load_GEFdata`.
     Loading data this way creates a `BGSystem` which also knows the appropriate units, constants and functions for this model.

    The `BaseGEF` is a compiled version of `.models.pai`. To compile other models, use the class factory `compile_model`.
    """
}



r"""
Welcome to our visitors tour of the **Gradient Expansion Formalism Factory**!

---

# GEFF: Established in 2025

This python package is designed to handle gauge-field production during cosmic inflation
using the *gradient expansion formalism* (GEF).

If you are interested in axion inflation, the package comes with everything you need:
- **a variety of flavors**
    - pure axion inflation (PAI) : *good ol' axion inflation*
    - fermionic axion inflation (FAI) : *axion inflation with Standard Model fermions!*
- **useful tools**
    - Resolve the dynamics of axion inflation including homogeneous backreaction.
    - Analyze the gauge-field spectrum.
    - Determine the vacuum and induced tensor power spectrum.
    - Compute gravitational-wave spectra.


But we don't want to hold you back! The package provides a flexible framework to create your **own GEF flavor**, with all built-in tools at your disposable. 
It is indeed a true GEF *factory* !

You can install this package using pip

```bash
pip install cosmo-geff
```

or using the `geff.yml` file found at the [https://github.com/riroro13/GEFF](GitHub repository) for this package,

```bash
conda env create -f geff.yml
``` 

If you use this package in your work, please cite (...).

---

# The refreshing taste of GEF

The GEF is a numerical technique to determine the dynamics and backreaction of gauge-fields during inflation
by directly evolving the time-dependent quantum expectation values of the gauge field, 
e.g., $\langle {\bf E}^2 \rangle$, $\langle {\bf B}^2 \rangle$, $\langle {\bf E} \cdot {\bf B} \rangle$ etc.
If this is the first time you encounter the GEF, here are some useful articles on the topic:
* [2109.01651](https://arxiv.org/abs/2109.01651)
* [2310.09186](https://arxiv.org/abs/2310.09186)
* [2408.16538](https://arxiv.org/abs/2408.16538)

The strategy behind the GEF is to take Maxwell's equations in an expanding spacetime,

<a name="max">$$\operatorname{div} {\bf E} = 0\, , \qquad \operatorname{div} {\bf B} = 0\, ,$$</a>
$$\dot{{\bf E}} + 2 H {\bf E} - \frac{1}{a}\operatorname{rot} {\bf B} + {\bf J} = 0 \, ,$$
$$\dot{{\bf B}}  + 2 H {\bf B} + \frac{1}{a}\operatorname{rot} {\bf E} = 0 \,$$

and use them to formulate a tower of ODEs for the quantities

$$ \mathcal{F}_\mathcal{E}^{(n)} = \frac{a^4}{k_{{\rm UV}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf E}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{a^2 k^{n+3}}{2 \pi^2 k_{{\rm UV}}^{n+4}}  \sum_{\lambda}\lambda^n |\dot{A}_\lambda(t,k)|^2\, ,$$
$$ \mathcal{F}_\mathcal{G}^{(n)} = -\frac{a^4}{2 k_{{\rm UV}}^{n+4}}\langle {\bf E} \cdot \operatorname{rot}^n {\bf B} + {\bf B} \cdot \operatorname{rot}^n {\bf E}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)} \frac{{\rm d} k}{k} \frac{a k^{n+4}}{2 \pi^2 k_{{\rm UV}}^{n+4}}\sum_{\lambda}\lambda^{n+1} \operatorname{Re}[\dot{A}_\lambda(t,k)A_\lambda^*(t,k)] \, ,$$
$$ \mathcal{F}_\mathcal{B}^{(n)} = \frac{a^4}{k_{{\rm UV}}^{n+4}}\langle {\bf B} \cdot \operatorname{rot}^n {\bf B}\rangle = \int\limits_{0}^{k_{{\rm UV}}(t)}\frac{{\rm d} k}{k} \frac{k^{n+5}}{2 \pi^{2}k_{{\rm UV}}^{n+4}} \sum_{\lambda}\lambda^n |A_\lambda(t,k)|^2 \, ,$$

where, $k_{\rm UV}$ is a suitably chosen UV regulator which can vary with time. 
For completeness, we have also given the expression for $\mathcal{F}_\mathcal{X}^{(n)}$ in terms of the mode functions $A_\lambda(t,k)$.

The ODE for the $\mathcal{F}_\mathcal{X}^{(n)}$'s are then given by

$$\frac{\rm d}{{\rm d} t} \mathcal{F}_\mathcal{E}^{(n)} + (4+n)\frac{{\rm d} \ln k_{\rm UV}}{{\rm d} t} \mathcal{F}_\mathcal{E}^{(n)}  + 2\frac{k_{\rm UV}}{a}\mathcal{F}_\mathcal{G}^{(n+1)} + 2 \frac{a^4}{k_{\rm UV}^{n+4}} \langle {\bf J} \cdot \operatorname{rot}^n {\bf E} \rangle =  S_{\mathcal{E}}^{(n)}\, , $$
$$\frac{\rm d}{{\rm d} t} \mathcal{F}_\mathcal{G}^{(n)} + (4+n)\frac{{\rm d} \ln k_{\rm UV}}{{\rm d} t} \mathcal{F}_\mathcal{G}^{(n)} - \frac{k_{\rm UV}}{a}\left(\mathcal{F}_\mathcal{E}^{(n+1)} - \mathcal{F}_\mathcal{B}^{(n+1)}\right) - \frac{a^4}{k_{\rm UV}^{n+4}} \langle {\bf J} \cdot \operatorname{rot}^n {\bf B} \rangle= S_{\mathcal{G}}^{(n)}\, , $$
$$\frac{\rm d}{{\rm d} t} \mathcal{F}_\mathcal{B}^{(n)} + (4+n)\frac{{\rm d} \ln k_{\rm UV}}{{\rm d} t} \mathcal{F}_\mathcal{B}^{(n)} - 2\frac{k_{\rm UV}}{a}\mathcal{F}_\mathcal{G}^{(n+1)}  =  S_{\mathcal{B}}^{(n)}\, .$$

Although these are infinitely many coupled ODE's, one can typically determine an analytical closing condition, such that they may be truncated at some order $n_{\rm tr}$.

The ODEs for the $\mathcal{F}_\mathcal{X}^{(n)}$ can now be simply solved alongside those of the inflationary background.
This way, one can handle gauge-field backreaction onto the background dynamics of inflation.

The GEFF package is designed to help the user in the process of solving these equations in the following way
 - pre-defined and ready-to-use [GEF flavors](#basics)
 - tailored [algorithm](#algorithm) to solve the inflationary background dynamics
 - options to [implement your own GEF flavor](#model_creation)

 ---

<a name="basics">
# Sampling the GEF flavors

GEF models come in various flavors, some of the most intresting applications of the GEF are already implemented in this package.
As the first part of our tour, we explore the basic application of the GEFF code using these pre-defined models.

## Choosing your flavor

We start by sampling a first flavor of our choice; the model "pai":

```python
from geff import compile_model

# Create the GEF model of our choice
paiGEF = compile_model("pai")
``` 
The object `paiGEF` is the compiled version of the model found under `.models.pai`. It defines an ODE solver 
which needs to be initialized with information on the setup we want to study:
The "pai" model expects information on the inflaton--vector coupling $\beta / M_{\rm P}$, 
initial conditions for the inflaton field, $\varphi(0)$, $\dot\varphi(0)$, and the shape of the inflaton potential, $V(\varphi)$.

In this example, we configure the model to start on the slow-roll attractor of a chaotic inflation potential:

$$ \varphi(0) = 15.55 M_{\rm P}, \qquad \dot{\varphi}(0) = -\sqrt{\frac{2}{3}} m M_{\rm P}, \qquad V(\varphi) = \frac{1}{2}m^2 \varphi^2$$

where $m = 6.16 \times 10^{-6} M_{\rm P}$. We set the coupling to $\beta = 15$.

The necessary information is passed to `paiGEF` as keyword arguments:
```python
import numpy as np

m = 6.16e-6
beta = 15

phi = 15.55
dphi = -np.sqrt(2/3)*m

def V(x): return 0.5*m**2*x**2
def dV(x): return m**2*x

mod = paiGEF(beta=beta, phi=phi, dphi=dphi, V=V, dV=dV)
```
If you want to know what input is expected by the GEF model, use the `print_input` method of `paiGEF`.

> **A note on units**:
> The pre-defined GEF flavors in the GEFF package work in Planck units $M_{\rm P}=1$. 
> From the input, the GEF determines the Hubble rate at initialization, $H_0$, (also in Planck units).
> Internally, the numerical routines work with dimensionless quantities, e.g., $\bar{X} = X H_0^{-a} M_{\rm P}^{-b}$ with $a$ and $b$ indicating
> how $X$ scales with an inverse timescale (e.g., $H_0$) and an energy scale (e.g., $M_{\rm P}$).
> For example, the dimensionless inflaton velocity would be like $\dot{\bar{\varphi}} = \dot{\varphi}/(H_0 M_{\rm P})$,
> i.e., $\dot{\varphi}$ scales like an amplitude, $\varphi \sim M_{\rm P}$, and a derivative, $\partial_t \sim H_0$.
> Don't worry, this is happening under the hood, but if you want more details, see `.bgtypes`.

## Getting a taste

Now, that our model is initialized, we can solve the inflationary background evolution from these initial conditions:
```python
sol, spec, info = mod.run()
```
The `run` method returned three objects. The evolution of the background dynamics is contained in `sol`,
the evolution on the gauge-field mode functions, $A_\lambda(t, k)$, are computed and returned as the
object `spec`, while `info` is just a byproduct that contains full information on the ODE solution in `sol`.
For basic applications, all information we actually want is in `sol` and `spec`.

Let us focus on `sol`. It is a `BGSystem` object which we can use to access the time evolution of several important inflationary quantities defined by our "pai" model.
For example, you can use it to make a basic plot showing the evolution of the energy densities during inflation as a function of $e$-folds

```python
import matplotlib.pyplot as plt

# Plot the evolution of the inflaton amplitude as a function of e-folds:
plt.plot(sol.N, sol.dphi**2/(6*sol.H**2)) # inflaton kinetic energy density
plt.plot(sol.N, (sol.E + sol.B)/(6*sol.H**2)) # electromagnetic energy density
plt.plot(sol.N, sol.V(sol.phi)/(3*sol.H**2)) # inflaton potential energy density

plt.yscale("log")
plt.ylim()

plt.show()
``` 

How did we know that `sol` owns the attributes `N`,`phi`, etc.? You can find out using `sol.value_names()`.
To print a full description of all available variables for "pai", you can use `paiGEF.print_ingredients()`. 
Otherwise, this information can also be found at `.models.pai`. 
If you need a brief description of any variable `X`, use `X.get_description()`.

> **A note on variables:**
> The variables encoded in a GEF model use a custom class called `Variable`. They work like a `numpy` array. 
> Additionally, constants are realized using the class `Constant`, and work like a `float`.
> Dimensionful functions like the inflaton potential use the `Func` class, and can be used like a regular function.
> These classes are defined to take care of unit conversions and are collectively attatched to a `BGSystem`. If you are curious, have a look at `.bgtypes`.

We did not only get back a `sol` object, but also `spec`. This object contains the gauge-field mode functions $A_\lambda(t,k)$.
The code computes these mode functions after having solved the dynamics of the background system ($H(t)$, $\varphi(t)$ etc.),
and uses them for estimating the convergence of the background solution. 
We briefly discuss this algorithm in [the next section](#algorithm).

The object `spec` is an instance of the `GaugeSpec` class, which defines useful methods for handling time-dependent gauge-field spectra.
For details, see `.mbm.GaugeSpec`.

> **A note on gauge-field spectra**
> If you are only interested in the background evolution and not in the gauge-field spectrum, set `nmodes=None` in `run()`.
> However, it is advised to compute gauge-field spectra to ensure consistency between the background evolution and the spectra.

Next, let us store our GEF and mode solutions in a file:
```python
# some dummy paths for illustration
gefpath = "some_gef_file.dat"
mbmpath = "some_mbm_file.dat"

sol.save_variables(gefpath)
spec.save_spec(mbmpath)
```

The data can be restored from these files using
```python
from geff import GaugeSpec

sol = mod.load_GEFdata(gefpath)
spec = GaugeSpec.read_spec(mbmpath)
```

> **A note on storage:**
> The `save_variables` method does not store information on constants or functions. So, to retrieve
> the full information on our GEF run, we need to use an appropriately configured instance of
> `paiGEF`. In the example above, this is achieved by reusing `mod`.

## A rich palette

Obtaining the inflationary dynamics is nice, but it is not all the GEFF can do. For example, let's use our results to compute the corresponding gravitational-wave spectrum:
```python
from geff.tools import PowSpecT, omega_gw

# Use sol to initialize the PowSpecT class
pt_fai = PowSpecT(sol)

# Compute the vacuum and induced power spectrum for 100 momentum modes k using spec:
ks, pt_spec = pt_fai.compute_pt(100, spec)

# from the power spectrum, we can then deduce the gravitational-wave spectrum:
f, gwspec = omega_gw(ks, pt_spec["tot"], Nend=sol.N[-1], Hend=sol.H[-1])
``` 
We can just use  `sol` to extract the relevant information on the background solution.
It's that easy!

To finish this first part of our tour, let us sample a second GEF flavor, "fai_kh":
```python
# initialize the model
faiGEF = GEFModel("fai_kh", {"picture":"electric"})

# chose initial conditions
mod = faiGEF(beta=...)

# solve the ODEs as before
sol, spec, info = mod.run(...)

...
...
```
This model comes in three varieties ("pictures") corresponding to the effective treatment of fermions in the model.
We specified the particular variety py passing a `settings` dictionary upon creating the model.

For more details on the available models, see `geff.models`.

<a name="algorithm">
# On the factory floor

Next on the tour of the GEF factory, we visit the production line.
The following diagram sketches the basic algorithm behind the `run` method.

```mermaid
---
config:
    flowchart:
        defaultRenderer: "elk"
    theme: 'base'
    themeVariables:
        secondaryColor: '#CFCFC6'
        tertiaryColor: '#FAFAFA'
---
graph TB
    subgraph A["`**GEFModel**`"]
        direction TB
        St((initial data <br>from model))--> GS
        subgraph GS["`**GEFSolver**`"]
            direction LR
            GS1[/initial data/] --> GS2[solve background ODEs] --> GS3[/background <br> dynamics/]
        end
        GS -.-> MbM
        subgraph MbM["`**ModeSolver**`"]
            direction LR
            MbM1[/background <br> dynamics/] --> MbM2[compute <br>mode functions] --> MbM3[/spectrum/]
        end
        MbM -.-> A2[compare <br>background dynamics <br>& spectrum]
        C[self-correction <br> using spectrum]
        GS --> A2
        A2 -.-> |disagreement| C -.->  GS
    A2 --> |agreement|Fin[finalize] --> R1((background <br> dynamics))
    Fin -.-> R2((spectrum))
    end
```

As we have seen in the [first section](#basics), the `run` method is executed as part of a GEF Model.
Each GEF Model consists of two components, the `GEFSolver` and the `ModeSolver`. The former determines the time-dependent inflationary background, 
the latter computes the mode functions $A_\lambda(t,k)$.
The two results can then be compared to eachother to assess the convergence of the background dynamics.
If the two disagree, the GEF will attempt to self-correct using $A_\lambda(t,k)$.

Note that, if you use `run(nmodes=None)`, the dotted lines in the diagram can be ignored; the background solution is immediately returned without
computing the gauge-field spectrum.

For more details on the `GEFSolver`, see `geff.solver`, while for the `ModeSolver` see `geff.mbm`.

---

<a name="model_creation">
# Create your own flavor

Having explored the potential of the GEFF code, you may be inclined to create your own GEF flavor.
To help you in this process, we show how to implement an example toy model.

> **Warning**: Before jumping into this section, we advise that you first familiarize yourself with the GEF method.
> Also, this tutorial works best, if you have a basic knowledge of the classes defined in `.bgtypes`.

## The first step is the hardest

First, we need to work out the mathematical formulation of our model.

Let us consider the case of Abelian gauge-field production in de Sitter space ($H={\rm const.}$) by a current of the type ${\bf J} = 2 H \xi {\bf B}$, 
where $\xi$ is a constant, which we refer to as instability parameter.The ODE tower for the gauge-field bilinears are then given by:

$$\frac{\rm d}{{\rm d} t} \mathcal{F}_\mathcal{E}^{(n)} + (4+n)\frac{{\rm d} \ln k_\mathrm{h}}{{\rm d} t} \mathcal{F}_\mathcal{E}^{(n)}  + 2\frac{k_\mathrm{h}}{a}\mathcal{F}_\mathcal{G}^{(n+1)} - 4 H \xi \mathcal{F}_\mathcal{G}^{(n)}=  S_{\mathcal{E}}^{(n)}\, , $$
$$\frac{{\rm d}}{{\rm d} t} \mathcal{F}_\mathcal{G}^{(n)} + (4+n)\frac{{\rm d} \ln k_\mathrm{h}}{{\rm d} t} \mathcal{F}_\mathcal{G}^{(n)} - \frac{k_\mathrm{h}}{a}\left(\mathcal{F}_\mathcal{E}^{(n+1)} - \mathcal{F}_\mathcal{B}^{(n+1)}\right) - 2 H \xi \mathcal{F}_\mathcal{B}^{(n)}= S_{\mathcal{G}}^{(n)}\, , $$
$$\frac{{\rm d}}{{\rm d} t} \mathcal{F}_\mathcal{B}^{(n)} + (4+n)\frac{{\rm d} \ln k_\mathrm{h}}{{\rm d} t} \mathcal{F}_\mathcal{B}^{(n)} - 2\frac{k_\mathrm{h}}{a}\mathcal{F}_\mathcal{G}^{(n+1)}  =  S_{\mathcal{B}}^{(n)}\, .$$

One can determine that a sensible regularization scale for this model is given by $k_{\rm h}(t) = 2aH\xi$. 

The boundary terms $S_\mathcal{X}^{(n)}$ are a consequence of the time dependence of $k_{\rm h}$. They are expressed in terms of Whittaker functions,
but the GEFF has a module that takes care of them. We will see this later.

Once we have worked out the equations, we can turn to writing a new model file.

## Only the best ingredients

The first thing we should state in our model file is the name of the new GEF flavor:
```python
import numpy as np

name = "tutorial"
```
> **A note on settings**: Following the model's name, you can also define a `settings` dictionary.
> This dictionary should contain some key&ndash;value pair defining its name and the default setting.
> The `settings` should be accompanied by a function called `interpret_settings`. This method will be called on model creation,
> so you can use it to define how user input settings are handled. See, e.g., `.models.SE_kh` for how this is done in practice.

Next, we need to define define and categorize the variables which appear in our model. This is taken care of by the functions `define_var`, `define_const`,  and `define_func`.
The `.bgtypes` module also contains some pre-defined variables, which are often encountered in GEF models.

There are three variables which every GEF model needs to define:
* $t$ - *cosmic time* : all ODE's are solved in terms of $t$ starting from $t=0$.
* $N$ - *$e$-folds* : needed by most internal methods. It should be defined such that $N(t=0)=0$.
* $H$ - *Hubble rate* : needed by most internal methods.

Beyond these staples, some extra variables appear in our ODE's:
* $\mathcal{F}_\mathcal{X}^{(n)}$ - *the gauge-field bilinears*
* $k_{\rm h}$ - *the UV regulator*
* $\xi$ -  *the instability parameter*

To properly account for the gauge field, we also need to add in the following three variables:
* $\mathcal{E}^{(0)} = \langle {\bf E}^2 \rangle$ - *called `E` by the GEF*
* $\mathcal{B}^{(0)} = \langle {\bf B}^2 \rangle$ - *called `B` by the GEF*
* $\mathcal{G}^{(0)} = -\langle {\bf E} \cdot {\bf B} \rangle$ - *called `G` by the GEF*

> **Note on gauge field bilinears** The time evolution of the variables $\mathcal{F}_\mathcal{X}^{(n)}$ will not be saved by the GEFF code,
> since we are typically only interest in the quantities with $n=0$. The output `info` returned by the `run` method contains the full information on $\mathcal{F}_\mathcal{X}^{(n)}$,
> but only the information on $n=0$ is passed to `sol` in the form of  `E`, `B` and `G`.

All these variables need to be defined in our model file. This is done as follows:
```python
from geff.bgtypes import define_const

# We make use of the fact that a lot of these variables are pre-defined:
from geff.bgtypes import t, N, a, E, B, G, kh, GF

# We also need to define some new objects:
H = define_const("H", qu_omega=1, qu_mu=0) # Hubble rate (scales like inverse time)
xi = define_const("xi", qu_omega=0, qu_mu=0) # instability parameter
```
We use `define_const` to define the constants for our model: $H$ and $\xi$.
The Hubble rate has mass dimension one, and scales with an inverse time-scale $\omega$ as, $H = \bar{H} \omega$.
It does not scale like an amplitude $\mu$. Hence, `qu_omega=1` and `qu_mu=0`.
Similarly, $\xi$ is just a number, and we set `qu_omega=0` and `qu_mu=0`.
This information needs to be passed to properly allow for unit conversions in the code.
More information on units and scaling is given in `.bgtypes`.

All the variables we have defined serve a specific purpose in our GEF model. To inform the GEFF of this, we need to classify each of them in one of these categories:
* **time**: a time variable (needs to be `t`)
* **dynamical**: variables whose time-evolution is determined from a differential equation.
* **gauge**: the gauge-field variable `GF` representing $\mathcal{F}_\mathcal{X}^{(n)}$.
* **static**: variables whose time-evolution is computed from other variables.
* **constant**: constants of time.
* **function**: functions of the above.

In our case, this would look as follows:
```python
quantities = {
            "time":[t], # this is mandatory!
            "dynamical":[kh], # kh is best evolved from an ODE
            "gauge":[GF], # state the obvious
            "static":[N, a, E, B, G], # directly computed from other variables
            "constant":[H, xi], # we assume de-Sitter space, and xi is a constant
            "function":[]  # our model does not need any functions
            }
```

## Write a recipe

With the variables defined, an important step towards our GEF model is already taken. 
Next, let us set up the `GEFSolver`. (For more details, see `.solver.GEFSolver`.)

Internally, the GEFF package uses `scipy.integrate.solve_ivp` to solve differential equations. 
This requires that ODE's are formulated as $\dot{\vec{y}} = f(t, \vec{y})$ with $\vec{y}$ as a `numpy` array.
However, the GEFF prefers the `BGSystem` class, which takes care of unit conversions.
So, we need to define how to translate between the two.

First up, we define how to interpret user input to initialize the array $\vec{y}$:
```python
def initial_conditions(sys, ntr):
    yini = np.zeros(1 + 3*(ntr+1))
    yini[0] = np.log(2*sys.xi*sys.H) #index 0 of yini is log(kh)

    # initialize all F_X^n as zero at indices [1:]

    return yini
``` 
Note how `sys` has attributes `xi` and `H`. These correspond to the variables we have defined in the previous step.

Importantly, upon defining `initial_conditions`, we also make a choice; which entries in $\vec{y}$ correspond to which dynamical variable.
In our simple toy model, there actually is no choice to be made: The GEFF package expects that the gauge-field variables are stored *after* all
dynamical variables, as the number of the $\mathcal{F}_\mathcal{X}^{(n)}$'s will vary depending on $n_{\rm tr}$.
Hence, $k_{\rm h}$ necessarily goes first. However, we do have the choice of evolving $\log k_{\rm h}$ instead of $k_{\rm h}$.


Next, we write the recipe used to define our GEF ODE. The ODE evolution is computed in two steps:
1. `update_values`: Update `sys` according to $\vec{y}(t)$.
2. `timestep`: Compute $\dot{\vec{y}}(t)$ from `sys`.

To define `update_values`, we can use all the variables which we have previously declared (they are assumed to be in numerical units):
```python
def update_values(t, y, sys):
    # evolution of spacetime
    sys.t.value = t
    sys.N.value = sys.t*sys.H #perfect de Sitter
    sys.a.value = np.exp(sys.N.value)

    # define how kh is computed from xi:
    sys.kh.value = np.exp(y[0]) # y[0] is log(kh)

    # use that y[1] = F_E^0, y[2] = F_B^0, y[3] = F_G^0
    rescale = (sys.kh/sys.a)**4
    sys.E.value = rescale * y[1]
    sys.B.value = rescale * y[2]
    sys.G.value = rescale * y[3]

    return
```
For `timestep`, we can make use of some pre-defined functions in the `.utility` module.

```python
from geff.utility.eom import gauge_field_ode
from geff.utility.boundary import boundary_approx
from geff.utility.general import heaviside

def timestep(t, y, sys):
    dydt = np.zeros_like(y)
    
    dlnkhdt = sys.H.value #dlnkhdt derivative
    dydt[0] = dlnkhdt

    xi = sys.xi.value

    # compute boundary terms
    W = boundary_approx(xi)

    # reshape arrays to fit gauge_field_ode
    Fcol = y[1:].shape[0]//3
    F = y[1:].reshape(Fcol,3)

    # compute the gauge-field ODEs
    dFdt = gauge_field_ode(F, sys.a, sys.kh, 2*sys.H*xi, W, dlnkhdt)
    # note that we can use 'a', 'kh' etc.;
    # 'update_values' is called before 'timestep'

    dydt[1:] = dFdt.reshape(Fcol*3) # reshape to fit dydt

    return dydt
``` 

These are all the ingredients we need to formulate the GEF ODE's. We can combine them using the `.solver.GEFSolver` class factory:

```python
from geff.solver import GEFSolver

solver = GEFSolver(initial_conditions, update_values, timestep, quantities)
``` 

> **Note**: This is just a basic `GEFSolver`. You can also define `Event` objects for a solver.
> An `Event` will check for a certain condition while the ODEs are being solved, and can terminate the solver if the condition is met.
> For example, you can define an `Event` to check if the end of inflation has been reached, or if some positive definite quantity has become negative.
> A `GEFSolver` can be configured to check for any `Event` occurrences and react to them in user-specified ways.
> For more details, see `geff.solver`.

We also should define the `ModeSolver`. In this toy model, we can just use a pre-defined class. For more complex situations, use `.mbm.ModeSolver`.
```python
from geff.mbm import BaseModeSolver

MbM = BaseModeSolver
```

## The finishing touch

 The last thing we need to do is define how our new GEF model is initialized.

First, we need to declare, what input our GEF model expects from the user.
There are two constants, $H$ and $\xi$, and the user should tell us their value:
```python
model_input = [H, xi, dxi]
``` 
Our model does not require other input; $\mathcal{F}_{\mathcal{X}}^{(n)}$ is initially set to zero, and $k_h(0)$ is determined from $\xi$ and $H$. 

The last step is to define the units of our GEF model based on the user input. This is achieved by the `define_units` function:
```python
def define_units(H):
    # The characteristic inverse time scale is the constant Hubble rate in Planck units
    freq = H
    # The charateristic energy scale is the Planck mass (in Planck units)
    amp = 1. 
    return freq, amp
```
The arguments of `define_units` necessarily needs to be a subset of `model_input`. In this case, we only need the Hubble rate, `H` to define our unit system:
The energy scale $\mu$ is the Planck mass in Planck units, while the inverse time scale is the constant Hubble rate $H$.

We are finally done! We can put everything we defined above in a file, which we call "tutorial.py", and we are good to go!

If all went well, you can now use your own GEF flavor just like the pre-defined ones:
```python
import numpy as np
from geff import compile_model
# Here, we assume you have saved your model as "tutorial.py"
import tutorial

TutorialGEF = compile_model(tutorial)

H = 5e-6
xi = 5

mod = TutorialGEF(H=H, xi=xi)

sol, spec, info = mod.run()
...
```
<script type="module">
  import mermaid from 'https://cdn.jsdelivr.net/npm/mermaid@11/dist/mermaid.esm.min.mjs';
  import elkLayouts from 'https://cdn.jsdelivr.net/npm/@mermaid-js/layout-elk@0/dist/mermaid-layout-elk.esm.min.mjs';
  mermaid.registerLayoutLoaders(elkLayouts);
</script>
"""
from .gef import compile_model
from .mbm import GaugeSpec
from .bgtypes import BGSystem

__version__ = "0.1"




"""
A module defining equations used to define a `GEFF.mbm.ModeSolver`.
"""
import numpy as np
from typing import Callable


def bd_classic(t: float, k : float) -> np.ndarray:
    r"""
    Returns gauge-field modes in Bunch&ndash;Davies vacuum:

    $$A_\lambda(k,t) \sim \frac{1}{\sqrt{2k}}exp{(-i \eta(t) k)}\, , \qquad -\eta(t) k \gg 1 \, .$$

    Parameters
    ----------
    t : float
        time of initialisation $t_{\rm init}$
    k : float
        comoving momentum $k$

    Returns
    -------
    y : NDArray
         $\sqrt{2k} A_\lambda(t_{\rm init},k)$ and $a\sqrt{2/k} \dot{A}_\lambda(t_{\rm init},k)$ in Bunch&ndash;Davies

    """
    return np.array([1., 0, 0, -1., 1, 0, 0, -1.])

def mode_equation_classic(t: float, y : np.ndarray, k : float, a : Callable,
                    xi : Callable, H : Callable) -> np.ndarray:
    r"""
    Mode equation for pure axion inflation:

    $$\ddot{A}_\lambda(t,k) + H \dot{A}_\lambda(t,k) + \left[\left(\frac{k}{a}\right)^2  - 2\lambda \left(\frac{k}{a}\right) \xi H \right]A_\lambda(t,k) = 0 \, .$$

    Parameters
    ----------
    t : float
        cosmic time $t$
    y : NDArray
        $\sqrt{2k} A_\lambda(t_{\rm init},k)$ and $a\sqrt{2/k} \dot{A}_\lambda(t_{\rm init},k)$
    k : float
        comoving momentum $k$
    a : Callable
        scale factor as function of time, $a(t)$
    xi : Callable
        instability parameter as function of time, $\xi(t)$
    H : Callable
        Hubble rate as function of time, $H(t)$
        
    Returns
    -------
    dydt : array
        an array of time derivatives of y
    """
    
    dydt = np.zeros_like(y)
    
    dis1 = k / a(t)
    dis2 = 2*H(t)*xi(t)

    #positive helicity
    lam = 1.
    #Real Part
    dydt[0] = y[1]*dis1
    dydt[1] = -(dis1  - lam * dis2) * y[0]
    
    #Imaginary Part
    dydt[2] = y[3]*dis1
    dydt[3] = -(dis1  - lam * dis2) * y[2]
    
    #negative helicity
    lam = -1.
    #Real Part
    dydt[4] = y[5]*dis1
    dydt[5] = -(dis1  - lam * dis2) * y[4]
    
    #Imaginary Part
    dydt[6] = y[7]*dis1
    dydt[7] = -(dis1  - lam * dis2) * y[6]
    
    return dydt

def damped_bd(t : float, k : float, a : Callable, delta : Callable,
              sigmaE : Callable) -> np.ndarray:
    r"""
    Returns gauge-field modes in damped Bunch&ndash;Davies vaccum:

    $$A_\lambda(k,t) \sim \sqrt{\frac{\Delta(t)}{2k}}exp{(-i \eta(t) k)}\, , \qquad -\eta(t) k \gg 1 \, .$$

    Parameters
    ----------
    t : float
        time of initialisation $t_{\rm init}$
    k : float
        comoving momentum $k$
    a : Callable
        scale factor as function of time, $a(t)$
    sigmaE : Callable
        electric damping, $\sigma_{\rm E}(t)$
    delta : Callable
        cumulative electric damping $\Delta(t) = \exp{\left(-\int \sigma_{\rm E}(t) {\rm d}t\right)}$

    Returns
    -------
    y : NDArray
        $\sqrt{2k} A_\lambda(t_{\rm init},k)$ and $a\sqrt{2/k} \dot{A}_\lambda(t_{\rm init},k)$ in Bunch&ndash;Davies

    """
    yini = np.array([1., -1/2*sigmaE(t)*a(t)/k, 0, -1.,
                     1., -1/2*sigmaE(t)*a(t)/k, 0, -1.])*np.sqrt( delta(t) )
    return yini


def mode_equation_SE_no_scale(t: float, y : np.ndarray, k : float,
                    a : Callable, xieff : Callable, H : Callable,
                    sigmaE : Callable) -> np.ndarray:
    r"""
    Mode equation for scale-dependent fermionic axion inflation:

    $$\ddot{A}_\lambda(t, k)  + \big( H + \sigma_{\rm E} \big) \dot{A}_\lambda(t, k) + \left[ \left(\frac{k}{a}\right)^2 - \lambda \frac{k}{a} \big( 2 \xi H + \sigma_{\rm B} \big) \right] A_\lambda(t, k) = 0\, .$$

    Parameters
    ----------
    t : float
        cosmic time $t$
    y : NDArray
        $\sqrt{2k} A_\lambda(t_{\rm init},k)$ and $a\sqrt{2/k} \dot{A}_\lambda(t_{\rm init},k)$
    k : float
        comoving momentum $k$
    a : Callable
        scale factor as function of time, $a(t)$
    xieff : Callable
        effective instability parameter as function of time, $\xi_{\rm eff}(t)$
    H : Callable
        Hubble rate as function of time, $H(t)$
    sigmaE : Callable
        electric damping, $\sigma_{\rm E}(t)$
        
    Returns
    -------
    dydt : array
        an array of time derivatives of y
    """
    
    dydt = np.zeros_like(y)
    
    drag = sigmaE(t)
    dis1 = k / a(t)
    dis2 = 2*H(t)*xieff(t)

    #positive helicity
    lam = 1.
    #Real Part
    dydt[0] = y[1]*dis1
    dydt[1] = - drag * y[1] - (dis1  - lam * dis2) * y[0]
    
    #Imaginary Part
    dydt[2] = y[3]*dis1
    dydt[3] = - drag * y[3] - (dis1  - lam * dis2) * y[2]
    
    #negative helicity
    lam = -1.
    #Real Part
    dydt[4] = y[5]*dis1
    dydt[5] = - drag * y[5] - (dis1  - lam * dis2) * y[4]
    
    #Imaginary Part
    dydt[6] = y[7]*dis1
    dydt[7] = - drag * y[7] - (dis1  - lam * dis2) * y[6]
    
    return dydt

def mode_equation_SE_scale(t: float, y : np.ndarray, k : float,
                    a : Callable, xi : Callable, H : Callable,
                    sigmaE : Callable, sigmaB : Callable,
                      kS : Callable) -> np.ndarray:
    r"""
    Mode equation for scale-independent fermionic axion inflation:

    $$\ddot{A}_\lambda(t, k)  + \big( H + \sigma_{\rm E} \Theta(t, k) \big) \dot{A}_\lambda(t, k) + \left[ \left(\frac{k}{a}\right)^2 - \lambda \frac{k}{a} \big( 2 \xi H + \sigma_{\rm B}\Theta(t, k) \big) \right] A_\lambda(t, k) = 0\, .$$

    Parameters
    ----------
    t : float
        cosmic time $t$
    y : NDArray
        $\sqrt{2k} A_\lambda(t_{\rm init},k)$ and $a\sqrt{2/k} \dot{A}_\lambda(t_{\rm init},k)$
    k : float
        comoving momentum $k$
    a : Callable
        scale factor as function of time, $a(t)$
    xi : Callable
        einstability parameter as function of time, $\xi(t)$
    H : Callable
        Hubble rate as function of time, $H(t)$
    sigmaE : Callable
        electric damping, $\sigma_{\rm E}(t)$
    sigmaB : Callable
        magnetic damping, $\sigma_{\rm B}(t)$
    kS : Callable
        no damping for $k > k_{\rm S}(t)$ from $\Theta(t, k)$
        
    Returns
    -------
    dydt : array
        an array of time derivatives of y
    """

    dydt = np.zeros_like(y)
    cut = max(a(t)*H(t), kS(t))
    theta = np.heaviside(cut - k, 0.5)
    
    drag = theta*sigmaE(t)
    dis1 = k / a(t)
    dis2 = 2*H(t)*xi(t) + theta*sigmaB(t)

    #positive helicity
    lam = 1.
    #Real Part
    dydt[0] = y[1]*dis1
    dydt[1] = - drag * y[1] - (dis1  - lam * dis2) * y[0]
    
    #Imaginary Part
    dydt[2] = y[3]*dis1
    dydt[3] = - drag * y[3] - (dis1  - lam * dis2) * y[2]
    
    #negative helicity
    lam = -1.
    #Real Part
    dydt[4] = y[5]*dis1
    dydt[5] = - drag * y[5] - (dis1  - lam * dis2) * y[4]
    
    #Imaginary Part
    dydt[6] = y[7]*dis1
    dydt[7] = - drag * y[7] - (dis1  - lam * dis2) * y[6]
    
    return dydt



